###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCb Bookkeeping database manager."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
import time

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

from LHCbDIRAC.BookkeepingSystem.Client.BaseESManager import BaseESManager
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.BookkeepingSystem.Client import objects
from LHCbDIRAC.BookkeepingSystem.Client.Help import Help

__RCSID__ = "$Id$"

INTERNAL_PATH_SEPARATOR = "/"

#############################################################################


class LHCbBookkeepingManager(BaseESManager):
  """creates the virtual file system."""

  __bookkeepingFolderProperties = ['name',
                                   'fullpath',
                                   ]
  # watch out for this ad hoc solution
  # if any changes made check all functions
  #
  __bookkeepingConfigurationPrefixes = ['ConfigName',  # configname
                                        'ConfigVersion',  # configversion
                                        'Simulation/DataTaking',
                                        'ProcessingPass',
                                        'EventType',  # event type
                                        'Production',  # production
                                        'FileType',  # file type
                                        ''
                                        ]

  __bookkeepingProductionPrefixes = ['PROD',
                                     'EVT',
                                     'FTY',
                                     ''
                                     ]

  __bookkeepingRunPrefixes = ['RUN',
                              'PAS',
                              'EVT',
                              'FTY',
                              ''
                              ]

  __bookkeepingEventtypePrefixes = ['ConfigName',
                                    'ConfigVersion',
                                    'EventType',
                                    'Simulation/DataTaking',
                                    'ProcessingPass',
                                    'Production',
                                    'FileType',
                                    '',
                                    ]
  __bookkeepingDatabasePrefixes = []

  __bookkeepingParameters = ['Configuration', 'Event type', 'Productions', 'Runlookup']

  __bookkeepingShortparameternames = {__bookkeepingParameters[0]: 'sim', __bookkeepingParameters[1]: 'evt',
                                      __bookkeepingParameters[2]: 'prod', __bookkeepingParameters[3]: 'run'}

  __bookkeepingQueryTypes = ['adv', 'std']

  #############################################################################
  def __init__(self, url=None, web=False, welcome=True):
    """initialize the values."""
    BaseESManager.__init__(self)
    self._BaseESManager___fileSeparator = INTERNAL_PATH_SEPARATOR
    # self.__pathSeparator = INTERNAL_PATH_SEPARATOR
    self.db_ = BookkeepingClient(url)
    if not web:
      self.fileCatalog = FileCatalog()
    self.helper_ = Help()

    self.__entityCache = {'/': (objects.Entity({'name': '/', 'fullpath': '/', 'expandable': True}), 0)}
    self.parameter_ = self.__bookkeepingParameters[0]
    self.files_ = []
    self.__filetypes = []
    self.comment = '#-- '

    self.treeLevels_ = -1
    self.advancedQuery_ = False
    if welcome:
      print('WELCOME')
      print("For more information use the 'help' command! ")
    self.dataQualities_ = {}

    retVal = self.db_.getAvailableFileTypes()
    if not retVal['OK']:
      gLogger.error(retVal)
    else:
      self.__filetypes = [i[0] for i in retVal['Value']['Records']]

  #############################################################################
  def setFileTypes(self, fileTypeList=list()):
    """it sets the file types.

    The parameter is a list of file type
    """
    if fileTypeList and len(fileTypeList) > 0:
      self.__filetypes = fileTypeList
    else:
      retVal = self.db_.getAvailableFileTypes()
      if not retVal['OK']:
        gLogger.error(retVal)
      else:
        self.__filetypes = [i[0] for i in retVal['Value']['Records']]

  #############################################################################
  def _updateTreeLevels(self, level):
    """tree level update."""
    self.treeLevels_ = level

  #############################################################################
  @staticmethod
  def setVerbose(value):
    """information printed."""
    objects.VERBOSE = value

  #############################################################################
  def setAdvancedQueries(self, value):
    """advanced queries."""
    self.advancedQuery_ = value

  #############################################################################
  def _getTreeLevels(self):
    """level of the current tree."""
    return self.treeLevels_

  #############################################################################
  def help(self):
    """help information."""
    if self.parameter_ == self.__bookkeepingParameters[0]:
      self.helper_.helpConfig(self._getTreeLevels())
    elif self.parameter_ == self.__bookkeepingParameters[1]:
      self.helper_.helpEventType(self._getTreeLevels())
    elif self.parameter_ == self.__bookkeepingParameters[2]:
      self.helper_.helpProcessing(self._getTreeLevels())

  #############################################################################
  def getPossibleParameters(self):
    """available parameters."""
    return self.__bookkeepingParameters

  #############################################################################
  def getCurrentParameter(self):
    """current parameters."""
    return self.__bookkeepingShortparameternames[self.parameter_]

  #############################################################################
  def getQueriesTypes(self):
    """types of queries."""
    if self.advancedQuery_:
      return self.__bookkeepingQueryTypes[0]
    else:
      return self.__bookkeepingQueryTypes[1]

  #############################################################################
  def setParameter(self, name):
    """query types."""
    if self.__bookkeepingParameters.__contains__(name):
      self.parameter_ = name
      self.treeLevels_ = -1
      if name == 'Configuration':
        self.__bookkeepingDatabasePrefixes = self.__bookkeepingConfigurationPrefixes
      elif name == 'Productions':
        self.__bookkeepingDatabasePrefixes = self.__bookkeepingProductionPrefixes
      elif name == 'Event type':
        self.__bookkeepingDatabasePrefixes = self.__bookkeepingEventtypePrefixes
      elif name == 'Runlookup':
        self.__bookkeepingDatabasePrefixes = self.__bookkeepingRunPrefixes
    else:
      gLogger.error("Wrong Parameter!")

  #############################################################################
  def getLogicalFiles(self):
    """lfn."""
    return self.files_

  #############################################################################
  def getFilesPFN(self):
    """pfn."""
    lfns = self.files_
    res = self.fileCatalog.getReplicas(lfns)
    return res

  #############################################################################
  def list(self, path="/", selectionDict=None, sortDict=None, startItem=0, maxitems=0):
    """list a path."""
    gLogger.debug(path)
    selectionDict = selectionDict if selectionDict is not None else {}
    sortDict = sortDict if sortDict is not None else {}

    if self.parameter_ == self.__bookkeepingParameters[0]:
      return self._listConfigs(path, sortDict, startItem, maxitems)
    elif self.parameter_ == self.__bookkeepingParameters[1]:
      return self._listEventTypes(path, sortDict, startItem, maxitems)
    elif self.parameter_ == self.__bookkeepingParameters[2]:
      return self._listProduction(path)
    elif self.parameter_ == self.__bookkeepingParameters[3]:
      return self._listRuns(path)

  #############################################################################
  def getLevelAndPath(self, path):
    """level and path."""
    if path == '/':
      return 0, [], ''  # it is the first level
    path = self.getAbsolutePath(path)['Value']  # shall we do this here or in the _processedPath()?
    processedPath = self._processPath(path)
    tmpPath = list(processedPath)
    if self.parameter_ == self.__bookkeepingParameters[1]:
      level, procpass = self.__getEvtLevel(tmpPath, [], level=0, start=False, end=False,
                                           processingpath='', startlevel=4)
    elif self.parameter_ == self.__bookkeepingParameters[3]:
      level, procpass = self.__getRunLevel(tmpPath, [], level=0, start=False, end=False,
                                           processingpath='', startlevel=1)
    else:
      level, procpass = self.__getLevel(path=tmpPath, visited=[],
                                        level=0, start=False,
                                        end=False, processingpath='',
                                        startlevel=3)
    self._updateTreeLevels(level)
    return level, processedPath, procpass

  #############################################################################
  # This method recursive visits all the tree nodes and found the processing pass
  def __getLevel(self, path, visited, level, start, end, processingpath, startlevel):
    """level."""
    for i in path:
      if level == startlevel and not start:
        for j in visited:
          path.remove(j)
        level += 1
        return self.__getLevel(path, visited,
                               level, start=True,
                               end=False, processingpath='',
                               startlevel=3)
      else:
        level += 1
        try:
          int(i)
        except ValueError as ex:
          gLogger.debug(str(self.__class__) + "__getLevel" + str(ex))
        else:
          if start:
            end = True
        if start and not end:
          level = startlevel
          processingpath += '/' + i
        elif end and level <= startlevel + 1:
          level = startlevel + 1
      visited += [i]
    return level, processingpath

  #############################################################################
  def __getRunLevel(self, path, visited, level, start, end, processingpath, startlevel):
    """run level."""
    for i in path:
      if level == startlevel and not start:
        for j in visited:
          path.remove(j)
        level += 1
        return self.__getRunLevel(path, visited,
                                  level, start=True,
                                  end=False,
                                  processingpath='',
                                  startlevel=1)
      else:
        level += 1
        try:
          int(i)
        except ValueError as ex:
          gLogger.warn(str(self.__class__) + "__getRunLevel" + str(ex))
        else:
          if start:
            end = True
        if start and not end:
          level = startlevel
          processingpath += '/' + i
        elif end and level <= startlevel + 1:
          level = startlevel + 1
      visited += [i]
    return level, processingpath

  #############################################################################
  # This method recursive visite all the tree nodes and found the processing pass
  def __getEvtLevel(self, path, visited, level, start, end, processingpath, startlevel):
    """evt level."""
    for i in path:
      if level == startlevel and not start:
        for j in visited:
          path.remove(j)
        level += 1
        return self.__getEvtLevel(path, visited,
                                  level, start=True,
                                  end=False, processingpath='',
                                  startlevel=4)
      else:
        level += 1
        try:
          int(i)
        except ValueError as ex:
          gLogger.warn(str(self.__class__) + "__getEvtLevel" + str(ex))
          if start and i in self.__filetypes:
            end = True
        else:
          if start:
            end = True
        if start and not end:
          level = startlevel
          processingpath += '/' + i
        elif end and level <= startlevel + 1:
          level = startlevel + 1
      visited += [i]
    return level, processingpath

  #############################################################################
  def _listConfigs(self, path, sortDict, startItem, maxitems):
    """list 1th tree."""
    entityList = list()
    levels, processedPath, procpass = self.getLevelAndPath(path)

    if levels == 0:  # configname
      self.clevelHeader_0()
      entityList += self.clevelBody_0(path, levels, )

    if levels == 1:  # config version
      in_dict = self.clevelHeader_1(processedPath)
      entityList += self.clevelBody_1(path, levels, in_dict)

    if levels == 2:  # sim or data desc
      in_dict = self.clevelHeader_2(processedPath)
      entityList += self.clevelBody_2(path, levels, in_dict)

    if levels == 3:  # processing
      in_dict = self.clevelHeader_3(processedPath)
      entityList += self.clevelBody_3(path, levels, in_dict, procpass)

    if levels == 4 and self.advancedQuery_:  # prod
      in_dict = self.clevelHeader_4(processedPath, procpass)
      entityList += self.clevelBody_4(path, levels, in_dict)
    elif levels == 4 and not self.advancedQuery_:
      processedPath += ['ALL']
      in_dict = self.clevelHeader_5(processedPath, procpass)
      entityList += self.clevelBody_5(path, 5, in_dict)

    if levels == 5 and self.advancedQuery_:  # file type
      in_dict = self.clevelHeader_5(processedPath, procpass)
      entityList += self.clevelBody_5(path, levels, in_dict)
    elif levels == 5 and not self.advancedQuery_:
      levels = 6

    if levels == 6 and startItem == 0 and maxitems == 0:  # files
      in_dict = self.clevelHeader_6(processedPath, procpass)
      entityList += self.clevelBody_6(path, levels, in_dict)
    elif levels == 6 and (startItem != 0 or maxitems != 0):  # files
      in_dict = self.clevelHeader_6(processedPath, procpass)
      entityList += self.clevelBodyLimited_6(path, levels, in_dict, sortDict, startItem, maxitems)
    return entityList

  #############################################################################
  def __addAll(self, path, levels, description):
    """level all."""
    if self.advancedQuery_:
      return self._getEntityFromPath(path, "ALL", levels, description)
    else:
      return None

  @staticmethod
  def __createPath(processedPath, name):
    """create a path."""
    path = ''
    for i in processedPath:
      string = '/' + i[0] + '_' + i[1]
      path += string

    path += '/' + name[0] + '_' + name[1]
    return path

  #############################################################################
  @staticmethod
  def clevelHeader_0():
    """configuration tree (first tree)"""
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Configurations names:")
    gLogger.debug("-----------------------------------------------------------")

    # list root
    gLogger.debug("listing Configuration Names")

  #############################################################################
  def clevelBody_0(self, path, levels):
    """1t level of the configuration tree."""
    entityList = list()
    result = self.db_.getAvailableConfigNames()

    if result['OK']:
      dbResult = result['Value']
      for record in dbResult['Records']:
        entityList += [self._getEntityFromPath(path, record[0], levels, None, {}, 'getAvailableConfigNames')]
      self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  @staticmethod
  def clevelHeader_1(processedPath):
    """second level."""
    gLogger.debug("listing configversions")
    in_dict = {'ConfigName': processedPath[0]}
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Configuration Name      | %s " % (processedPath[0]))

    gLogger.debug("Available Config Versions:")
    return in_dict

  def clevelBody_1(self, path, levels, in_dict):
    """second."""
    entityList = list()
    result = self.db_.getConfigVersions(in_dict)
    if result['OK']:
      dbResult = result['Value']
      description = dbResult["ParameterNames"][0]
      for record in dbResult['Records']:
        entityList += [self._getEntityFromPath(path, record[0], levels, description, in_dict, 'getConfigVersions')]
      self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  @staticmethod
  def clevelHeader_2(processedPath):
    """third."""
    gLogger.debug("listing Simulation Conditions!")
    in_dict = {'ConfigName': processedPath[0],
               'ConfigVersion': processedPath[1]}

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Available Simulation Conditions:")
    return in_dict

  #############################################################################
  def clevelBody_2(self, path, levels, in_dict):
    """build the tree node."""
    entityList = list()
    result = self.db_.getConditions(in_dict)
    if result['OK']:
      dbResult = result['Value']
      if dbResult[0]["TotalRecords"] > 0:
        add = self.__addAll(path, levels, 'Simulation Conditions/DataTaking')
        if add:
          entityList += [add]
        for record in dbResult[0]['Records']:
          value = {}
          j = 0
          for i in dbResult[0]['ParameterNames']:
            value[i] = record[j]
            j += 1
          entityList += [self._getSpecificEntityFromPath(path, value,
                                                         record[1], levels,
                                                         None, 'Simulation Conditions/DataTaking',
                                                         in_dict, 'getConditions')]
        self._cacheIt(entityList)
      if dbResult[1]["TotalRecords"] > 0:
        for record in dbResult[1]['Records']:
          value = {}
          j = 0
          for i in dbResult[1]['ParameterNames']:
            value[i] = record[j]
            j += 1
          entityList += [self._getSpecificEntityFromPath(path, value, record[1], levels,
                                                         None, 'Simulation Conditions/DataTaking',
                                                         in_dict, 'getConditions')]
        self._cacheIt(entityList)

    else:
      gLogger.error(result['Message'])

    return entityList

  #############################################################################
  @staticmethod
  def clevelHeader_3(processedPath):
    """fourth level."""
    gLogger.debug("listing processing pass")
    in_dict = {'ConfigName': processedPath[0],
               'ConfigVersion': processedPath[1],
               'ConditionDescription': processedPath[2]}

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)
    gLogger.debug("-----------------------------------------------------------")

    gLogger.debug("Available processing pass:\n")
    return in_dict

  #############################################################################
  def clevelBody_3(self, path, levels, in_dict, procpass):
    """fourth tree node."""
    entityList = list()
    in_dict['ProcessingPass'] = procpass
    result = self.db_.getProcessingPass(in_dict, procpass)
    if result['OK']:
      dbResult = result['Value']
      if dbResult[0]['TotalRecords'] > 0:  # it is a processing pass
        add = self.__addAll(path, levels, 'Processing Pass')
        if add:
          entityList += [add]
        for record in dbResult[0]['Records']:
          entityList += [self._getEntityFromPath(path, record[0],
                                                 levels, 'Processing Pass',
                                                 in_dict, 'getProcessingPass')]
        self._cacheIt(entityList)
      if dbResult[1]['TotalRecords'] > 0:
        value = {}
        for record in dbResult[1]['Records']:
          value = {'Event Type': record[0], 'Description': record[1]}
          entityList += [self._getSpecificEntityFromPath(path, value, str(record[0]),
                                                         levels, None, 'Event types',
                                                         in_dict, 'getProcessingPass')]
        self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  @staticmethod
  def clevelHeader_4(processedPath, procpass):
    """5th level."""
    gLogger.debug("listing event types")
    retVal = procpass.split('/')[1:]
    for i in retVal:
      processedPath.remove(i)

    in_dict = {'ConfigName': processedPath[0],
               'ConfigVersion': processedPath[1],
               'ConditionDescription': processedPath[2],
               'EventType': processedPath[3]}
    in_dict['ProcessingPass'] = procpass

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters: ")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)
    gLogger.debug("-----------------------------------------------------------")

    gLogger.debug("Available event types types:")
    return in_dict

  #############################################################################
  def clevelBody_4(self, path, levels, in_dict):
    """5th tree node."""
    entityList = list()
    result = self.db_.getProductions(in_dict)
    if result['OK']:
      dbResult = result['Value']
      for record in dbResult['Records']:
        entityList += [self._getEntityFromPath(path, str(record[0]), levels,
                                               'Production(s)/Run(s)', in_dict, 'getProductions')]
      self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  def clevelHeader_5(self, processedPath, procpass):
    """6th tree."""
    gLogger.debug("listing event types")
    retVal = procpass.split('/')[1:]
    for i in retVal:
      processedPath.remove(i)

    if self.advancedQuery_:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'ConditionDescription': processedPath[2],
                 'EventType': processedPath[3],
                 'Production': processedPath[4]}
    else:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'ConditionDescription': processedPath[2],
                 'EventType': processedPath[3]}
    in_dict['ProcessingPass'] = procpass

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters: ")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)
    gLogger.debug("-----------------------------------------------------------")

    gLogger.debug("Available event types types:")
    return in_dict

  #############################################################################
  def clevelBody_5(self, path, levels, in_dict):
    """build the 6th tree node."""
    entityList = list()
    in_dict['Visible'] = 'Y'
    result = self.db_.getFileTypes(in_dict)
    if result['OK']:
      dbResult = result['Value']
      for record in dbResult['Records']:
        entityList += [self._getEntityFromPath(path, record[0], levels, 'FileTypes', in_dict, 'getFileTypes')]
      self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  def clevelHeader_6(self, processedPath, procpass):
    """7th tree prepare."""
    gLogger.debug("listing event types")

    retVal = procpass.split('/')[1:]
    for i in retVal:
      processedPath.remove(i)

    if self.advancedQuery_:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'ConditionDescription': processedPath[2],
                 'EventType': processedPath[3],
                 'Production': processedPath[4],
                 'FileType': processedPath[5]}
    else:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'ConditionDescription': processedPath[2],
                 'EventType': processedPath[3],
                 'FileType': processedPath[4]}
    in_dict['ProcessingPass'] = procpass

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters: ")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)
    gLogger.debug("-----------------------------------------------------------")

    return in_dict

  #############################################################################
  def clevelBody_6(self, path, levels, in_dict):
    """build the 7th tree."""
    entityList = list()
    in_dict['DataQuality'] = self.__getSelectedQualities()
    in_dict['Visible'] = 'Y'
    result = self.db_.getFilesWithMetadata(in_dict)
    if result['OK']:
      for record in result['Value']['Records']:
        value = {'name': record[0], 'EventStat': record[1], 'FileSize': record[2], 'CreationDate': record[3],
                 'JobStart': record[4], 'JobEnd': record[5], 'WorkerNode': record[6],
                 'FileType': in_dict['FileType'], 'RunNumber': record[8],
                 'FillNumber': record[9], 'FullStat': record[10], 'DataqualityFlag': record[11],
                 'EventType': in_dict['EventType'],
                 'EventInputStat': record[12], 'TotalLuminosity': record[13], 'Luminosity': record[14],
                 'InstLuminosity': record[15], 'TCK': record[16]}
        self.files_ += [record[0]]
        entityList += [self._getEntityFromPath(path, value, levels, 'List of files', in_dict, 'getFilesWithMetadata')]
      self._cacheIt(entityList)
    else:
      return result
    return entityList

  #############################################################################
  def clevelBodyLimited_6(self, path, levels, in_dict, sortDict, startItem, maxitems):
    """7th tree node for web."""
    entityList = list()
    in_dict['DataQuality'] = self.__getSelectedQualities()
    result = self.__getFiles(in_dict, sortDict, startItem, maxitems)
    for record in result['Records']:
      value = {'name': record[0], 'EventStat': record[1], 'FileSize': record[2], 'CreationDate': record[3],
               'JobStart': record[4], 'JobEnd': record[5], 'WorkerNode': record[6],
               'FileType': in_dict['FileType'], 'EventType': in_dict['EventType'],
               'RunNumber': record[9], 'FillNumber': record[10], 'FullStat': record[11],
               'DataqualityFlag': record[12], 'EventInputStat': record[13],
               'TotalLuminosity': record[14], 'Luminosity': record[15], 'InstLuminosity': record[16], 'TCK': record[17]}

      self.files_ += [record[0]]
      entityList += [self._getEntityFromPath(path, value, levels, 'List of files', in_dict, 'getFilesWithMetadata')]
    self._cacheIt(entityList)
    return entityList

  #############################################################################
  def _listEventTypes(self, path, sortDict, startItem, maxitems):
    """second tree based on event type."""
    entityList = list()
    levels, processedPath, procpass = self.getLevelAndPath(path)

    if levels == 0:  # configname
      self.clevelHeader_0()
      entityList += self.clevelBody_0(path, levels, )

    if levels == 1:  # config version
      in_dict = self.clevelHeader_1(processedPath)
      entityList += self.clevelBody_1(path, levels, in_dict)

    if levels == 2:  # event type
      in_dict = self.clevelHeader_2(processedPath)
      entityList += self.elevelBody_2(path, levels, in_dict)

    if levels == 3:  # sim ot daq desq
      in_dict = self.elevelHeader_3(processedPath)
      entityList += self.elevelBody_3(path, levels, in_dict)

    if levels == 4:  # processing pass
      in_dict = self.elevelHeader_4(processedPath)
      entityList += self.elevelBody_4(path, levels, in_dict, procpass)

    if self.advancedQuery_ and levels == 5:
      in_dict = self.elevelHedaer_5(processedPath, procpass)
      entityList += self.clevelBody_5(path, levels, in_dict)
    elif levels == 5:
      levels = 6

    if levels == 6 and startItem == 0 and maxitems == 0:  # files
      in_dict = self.elevelHeader_6(processedPath, procpass)
      entityList += self.clevelBody_6(path, levels, in_dict)
    elif levels == 6 and (startItem != 0 or maxitems != 0):  # files
      in_dict = self.elevelHeader_6(processedPath, procpass)
      entityList += self.clevelBodyLimited_6(path, levels, in_dict, sortDict, startItem, maxitems)

    return entityList

  #############################################################################
  def elevelBody_2(self, path, levels, in_dict):
    """event type based queries."""
    entityList = list()
    result = self.db_.getEventTypes(in_dict)
    if result['OK']:
      dbResult = result['Value']
      if len(dbResult) > 1:
        add = self.__addAll(path, levels, 'Event types')
        if add:
          entityList += [add]
      for record in dbResult['Records']:
        value = {'Event Type': record[0], 'Description': record[1]}
        entityList += [self._getSpecificEntityFromPath(path, value, str(record[0]),
                                                       levels, None, 'Event types',
                                                       in_dict, 'getEventTypes')]
      self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
      return result
    return entityList

  #############################################################################
  @staticmethod
  def elevelHeader_3(processedPath):
    """event type based tree node."""
    gLogger.debug("listing simulation conditions")

    in_dict = {'ConfigName': processedPath[0], 'ConfigVersion': processedPath[1], 'EventType': processedPath[2]}

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)
    gLogger.debug("-----------------------------------------------------------")

    gLogger.debug("Available conditions:")
    return in_dict

  #############################################################################
  def elevelBody_3(self, path, levels, in_dict):
    """building the tree node."""

    entityList = list()

    result = self.db_.getConditions(in_dict)
    if result['OK']:
      dbResult = result['Value']
      if dbResult[0]["TotalRecords"] > 0:
        add = self.__addAll(path, levels, 'Simulation Conditions/DataTaking')
        if add:
          entityList += [add]
        for record in dbResult[0]['Records']:
          value = {}
          j = 0
          for i in dbResult[0]['ParameterNames']:
            value[i] = record[j]
            j += 1
          entityList += [self._getSpecificEntityFromPath(path, value, record[1],
                                                         levels, None, 'Simulation Conditions/DataTaking',
                                                         in_dict, 'getConditions')]
        self._cacheIt(entityList)
      if dbResult[1]["TotalRecords"] > 0:
        for record in dbResult[1]['Records']:
          value = {}
          j = 0
          for i in dbResult[1]['ParameterNames']:
            value[i] = record[j]
            j += 1
          entityList += [self._getSpecificEntityFromPath(path, value, record[1], levels,
                                                         None, 'Simulation Conditions/DataTaking',
                                                         in_dict, 'getConditions')]
        self._cacheIt(entityList)

    else:
      gLogger.error(result['Message'])

    return entityList

  #############################################################################
  @staticmethod
  def elevelHeader_4(processedPath):
    """prepare the 5th query."""
    gLogger.debug("listing processing pass")

    in_dict = {'ConfigName': processedPath[0],
               'ConfigVersion': processedPath[1],
               'EventType': processedPath[2],
               'ConditionDescription': processedPath[3]
               }

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters: ")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)
    gLogger.debug("-----------------------------------------------------------")

    gLogger.debug("Available processing pass types:")
    return in_dict

  #############################################################################
  def elevelBody_4(self, path, levels, in_dict, procpass):
    """make the tree node."""
    entityList = list()

    result = self.db_.getProcessingPass(in_dict, procpass)
    if result['OK']:
      dbResult = result['Value']
      if dbResult[0]['TotalRecords'] > 0:  # it is a processing pass
        add = self.__addAll(path, levels, 'Processing Pass')
        if add:
          entityList += [add]
        for record in dbResult[0]['Records']:
          entityList += [self._getEntityFromPath(path, record[0],
                                                 levels, 'Processing Pass',
                                                 in_dict, 'getProcessingPass')]
        self._cacheIt(entityList)
      elif self.advancedQuery_:
        in_dict['ProcessingPass'] = procpass
        result = self.db_.getProductions(in_dict)
        if result['OK']:
          dbResult = result['Value']
          for record in dbResult['Records']:
            entityList += [self._getEntityFromPath(path, str(record[0]),
                                                   levels, 'Production(s)/Run(s)',
                                                   in_dict, 'getProductions')]
          self._cacheIt(entityList)

    if len(procpass) > 0:
      in_dict['ProcessingPass'] = procpass
      in_dict['Visible'] = 'Y'
      result = self.db_.getFileTypes(in_dict)
      if result['OK']:
        dbResult = result['Value']
        for record in dbResult['Records']:
          entityList += [self._getEntityFromPath(path, record[0], levels, 'FileTypes', in_dict, 'getFileTypes')]
        self._cacheIt(entityList)
      else:
        gLogger.error(result['Message'])
    return entityList

  #############################################################################
  def elevelHedaer_5(self, processedPath, procpass):
    """prepare tree node."""
    retVal = procpass.split('/')[1:]
    for i in retVal:
      processedPath.remove(i)

    if self.advancedQuery_:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'EventType': processedPath[2],
                 'ConditionDescription': processedPath[3],
                 'Production': processedPath[4]}
    else:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'EventType': processedPath[2],
                 'ConditionDescription': processedPath[3]}
    in_dict['ProcessingPass'] = procpass

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:   ")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Available Production(s):")
    return in_dict

  #############################################################################
  def elevelHeader_6(self, processedPath, procpass):
    """prepare tree mode."""
    retVal = procpass.split('/')[1:]
    for i in retVal:
      processedPath.remove(i)

    if self.advancedQuery_:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'EventType': processedPath[2],
                 'ConditionDescription': processedPath[3],
                 'Production': processedPath[4],
                 'FileType': processedPath[5]}
    else:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'EventType': processedPath[2],
                 'ConditionDescription': processedPath[3],
                 'FileType': processedPath[4]}
    in_dict['ProcessingPass'] = procpass

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:   ")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Available file types:")
    return in_dict

  #############################################################################
  def _listProduction(self, path):
    """production lookup."""
    entityList = list()
    levels, processedPath, procpass = self.getLevelAndPath(path)
    gLogger.debug(str(procpass))

    if levels == 0:
      self.plevelHeader_0()
      entityList += self.plevelBody_0(path, levels)

    if levels == 1:
      in_dict = self.plevelHeader_2(processedPath)
      entityList += self.plevelBody_2(path, levels, in_dict)

    if levels == 2:
      in_dict = self.plevelHeader_3(processedPath)
      entityList += self.plevelBody_3(path, levels, in_dict)

    if levels == 3:
      in_dict = self.plevelHeader_4(processedPath)
      entityList += self.plevelBody_4(path, levels, in_dict)

    return entityList

  #############################################################################
  def _listRuns(self, path):
    """run lookup."""
    entityList = list()

    levels, processedPath, procpass = self.getLevelAndPath(path)

    if levels == 0:
      self.rlevelHeader_0()
      entityList += self.rlevelBody_0(path, levels)

    if levels == 1:
      in_dict = self.rlevelHeader_2(processedPath)
      entityList += self.rlevelBody_2(path, levels, in_dict, procpass)

    if levels == 2:
      in_dict = self.rlevelHeader_3(processedPath, procpass)
      entityList += self.rlevelBody_3(path, levels, in_dict)

    if levels == 3:
      in_dict = self.rlevelHeader_4(processedPath, procpass)
      entityList += self.clevelBody_6(path, levels, in_dict)

    return entityList

  #############################################################################
  @staticmethod
  def plevelHeader_0():
    """prepare production lookup tree node."""
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("productions:")
    gLogger.debug("-----------------------------------------------------------")

    # list root
    gLogger.debug("listing productions")

  #############################################################################
  def plevelBody_0(self, path, levels):
    """make the node of the production lookup tree."""
    entityList = list()
    result = self.db_.getAvailableProductions()

    if result['OK']:
      dbResult = result['Value']
      for record in dbResult:
        prod = str(record[0])
        entityList += [self._getEntityFromPath(path, str(prod), levels, 'Production(s)/Run(s)')]
      self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  @staticmethod
  def plevelHeader_2(processedPath):
    """prepare the tree node."""
    gLogger.debug("listing eventtype")

    in_dict = {'Production': processedPath[0]}

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)

    gLogger.debug("Available event types:")
    return in_dict

  #############################################################################
  def plevelBody_2(self, path, levels, in_dict):
    """make the tree node."""
    entityList = list()

    result = self.db_.getEventTypes(in_dict)
    if result['OK']:
      dbResult = result['Value']
      if len(dbResult) > 1:
        add = self.__addAll(path, levels, 'Event types')
        if add:
          entityList += [add]
      for record in dbResult['Records']:
        value = {'Event Type': record[0], 'Description': record[1]}
        entityList += [self._getSpecificEntityFromPath(path, value,
                                                       str(record[0]),
                                                       levels, None,
                                                       'Event types',
                                                       in_dict, 'getEventTypes')]
      self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  @staticmethod
  def plevelHeader_3(processedPath):
    """prepare tree node."""
    gLogger.debug("listing file types")
    in_dict = {'Production': processedPath[0], 'EventType': processedPath[1]}
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)

    gLogger.debug("Available file types:")
    return in_dict

  #############################################################################
  def plevelBody_3(self, path, levels, in_dict):
    """make tree node."""
    entityList = list()
    in_dict['Visible'] = 'Y'
    result = self.db_.getFileTypes(in_dict)
    if result['OK']:
      dbResult = result['Value']
      for record in dbResult['Records']:
        entityList += [self._getEntityFromPath(path, record[0], levels, 'FileTypes', in_dict, 'getFileTypes')]
      self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  @staticmethod
  def plevelHeader_4(processedPath):
    """prepare the tree node."""
    gLogger.debug("listing file types")
    in_dict = {'Production': processedPath[0], 'EventType': processedPath[1], 'FileType': processedPath[2]}

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)

    gLogger.debug("Available files:")
    return in_dict

  #############################################################################
  def plevelBody_4(self, path, levels, in_dict):
    """make the tree node."""
    entityList = list()
    in_dict['DataQuality'] = self.__getSelectedQualities()
    in_dict['Visible'] = 'Y'
    result = self.db_.getFilesWithMetadata(in_dict)
    if result['OK']:
      for record in result['Value']['Records']:
        value = {'name': record[0], 'EventStat': record[1], 'FileSize': record[2],
                 'CreationDate': record[3], 'JobStart': record[4], 'JobEnd': record[5],
                 'WorkerNode': record[6], 'FileType': in_dict['FileType'], 'RunNumber': record[8],
                 'FillNumber': record[9], 'FullStat': record[10], 'DataqualityFlag': record[11],
                 'EventType': in_dict['EventType'], 'EventInputStat': record[12],
                 'TotalLuminosity': record[13], 'Luminosity': record[14],
                 'InstLuminosity': record[15], 'TCK': record[16]}
        self.files_ += [record[0]]
        entityList += [self._getEntityFromPath(path, value,
                                               levels, 'List of files',
                                               in_dict, 'getFilesWithMetadata')]
      self._cacheIt(entityList)
    else:
      return result

    return entityList

  @staticmethod
  def rlevelHeader_0():
    """prepare run lookup node."""
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Runs:")
    gLogger.debug("-----------------------------------------------------------")

    # list root
    gLogger.debug("listing runs")

  #############################################################################
  def rlevelBody_0(self, path, levels):
    """make tree node."""
    entityList = list()
    result = self.db_.getAvailableRuns()

    if result['OK']:
      dbResult = result['Value']
      for record in dbResult:
        run = str(record[0])
        entityList += [self._getEntityFromPath(path, str(run), levels, 'Run(s)')]
      self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  @staticmethod
  def rlevelHeader_2(processedPath):
    """prepare tree node."""
    gLogger.debug("listing processing pass")
    in_dict = {'RunNumber': processedPath[0]}
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)

    gLogger.debug("Available processing pass:")
    return in_dict

  #############################################################################
  def rlevelBody_2(self, path, levels, in_dict, procpass):
    """make tree node."""
    entityList = list()
    in_dict['ProcessingPass'] = procpass
    result = self.db_.getProcessingPass(in_dict, procpass)
    if result['OK']:
      dbResult = result['Value']
      if dbResult[0]['TotalRecords'] > 0:  # it is a processing pass
        add = self.__addAll(path, levels, 'Processing Pass')
        if add:
          entityList += [add]
      for record in dbResult[0]['Records']:
        entityList += [self._getEntityFromPath(path, record[0],
                                               levels, 'Processing Pass',
                                               in_dict, 'getProcessingPass')]
      self._cacheIt(entityList)
      if dbResult[1]['TotalRecords'] > 0:
        value = {}
        for record in dbResult[1]['Records']:
          value = {'Event Type': record[0], 'Description': record[1]}
          entityList += [self._getSpecificEntityFromPath(path, value, str(record[0]),
                                                         levels, None, 'Event types',
                                                         in_dict, 'getProcessingPass')]
        self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  @staticmethod
  def rlevelHeader_3(processedPath, procpass):
    """prepare node of the run lookup tree."""
    gLogger.debug("listing eventtypes")
    retVal = procpass.split('/')[1:]
    for i in retVal:
      processedPath.remove(i)

    in_dict = {'RunNumber': processedPath[0], 'EventType': processedPath[1]}
    in_dict['ProcessingPass'] = procpass

    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)

    gLogger.debug("Available eventtypes types:")
    return in_dict

  #############################################################################
  def rlevelBody_3(self, path, levels, in_dict):
    """make tree node."""
    entityList = list()
    in_dict['Visible'] = 'Y'
    result = self.db_.getFileTypes(in_dict)
    if result['OK']:
      dbResult = result['Value']
      for record in dbResult['Records']:
        entityList += [self._getEntityFromPath(path, record[0], levels, 'FileTypes', in_dict, 'getFileTypes')]
      self._cacheIt(entityList)
    else:
      gLogger.error(result['Message'])
    return entityList

  #############################################################################
  @staticmethod
  def rlevelHeader_4(processedPath, procpass):
    """prepare tree."""
    gLogger.debug("listing file types")
    retVal = procpass.split('/')[1:]
    for i in retVal:
      processedPath.remove(i)

    in_dict = {'RunNumber': processedPath[0], 'EventType': processedPath[1], 'FileType': processedPath[2]}
    in_dict['ProcessingPass'] = procpass
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug("Selected parameters:")
    gLogger.debug("-----------------------------------------------------------")
    gLogger.debug(in_dict)

    gLogger.debug("Available file types:")
    return in_dict

  #############################################################################
  @staticmethod
  def _getEntityFromPath(presentPath, newPathElement, level, leveldescription=None, selection=None, method=None):
    """create a entity."""
    if isinstance(newPathElement, dict):
      # this must be a file
      entity = objects.Entity(newPathElement)
      newPathElement = str(entity['name']).rsplit("/", 1)[1]
      entity.update({'FileName': entity['name']})
      expandable = False
      entity.update({'expandable': expandable})
      if selection is not None:
        entity.update({'selection': selection})

      if method is not None:
        entity.update({'method': method})

    else:
      # this must be a folder
      entity = objects.Entity()
      name = newPathElement

      expandable = True

      fullPath = presentPath.rstrip(INTERNAL_PATH_SEPARATOR)
      fullPath += INTERNAL_PATH_SEPARATOR + \
          name

      entity.update({'name': name, 'fullpath': fullPath, 'expandable': expandable})

      if leveldescription is not None:
        entity.update({'level': leveldescription})

      if leveldescription == 'FileTypes':
        entity.update({'showFiles': 0})

      if selection is not None:
        entity.update({'selection': selection})

      if method is not None:
        entity.update({'method': method})

      elif level == 5:
        entity.update({'showFiles': 0})
    return entity

  #############################################################################
  @staticmethod
  def _getSpecificEntityFromPath(presentPath, value,
                                 newPathElement, level,
                                 description=None,
                                 leveldescription=None,
                                 selection=None, method=None):
    """crate a specific entity."""
    if isinstance(value, dict):
      entity = objects.Entity(value)
      name = newPathElement

      expandable = True

      fullPath = presentPath.rstrip(INTERNAL_PATH_SEPARATOR)
      fullPath += INTERNAL_PATH_SEPARATOR + \
          name

      if description is not None:
        entity.update({'name': description, 'fullpath': fullPath, 'expandable': expandable})
      else:
        entity.update({'name': name, 'fullpath': fullPath, 'expandable': expandable})
      if leveldescription is not None:
        entity.update({'level': leveldescription})

      if selection is not None:
        entity.update({'selection': selection})

      if method is not None:
        entity.update({'method': method})

      if level == 5:
        entity.update({'showFiles': 0})

    return entity

  #############################################################################
  def _processPath(self, path):
    """takes an absolute path and returns of tuples with prefixes and posfixes
    of path elements.

    If invalid path returns null
    """
    path = path.strip(INTERNAL_PATH_SEPARATOR + " ")
    paths = path.split(self.getPathSeparator())
    return paths

  #############################################################################
  def _cacheIt(self, entityList):
    """it caches an entity or a list of entities."""
    if isinstance(entityList, objects.Entity):
      # convert it into a list
      entityList = [entityList]
    elif not isinstance(entityList, list):
      # neither entity nor list
      gLogger.warn("couldn't cache invalid entity(list) of type " + str(entityList.__class__))
      return

    for entity in entityList:
      # TO IMPLEMENT!! time of the caching
      try:
        if 'fullpath' in entity:
          self.__entityCache.update({entity['fullpath']: (entity, 0)})
      except ValueError as ex:
        gLogger.warn("couldn't cache entity(?) " + str(entity) + "  " + str(ex))
        return S_ERROR('couldnt cache entity!')

  #############################################################################
  @staticmethod
  def getAbsolutePath(path):
    """get current working directory if empty."""
    if path in ["", ".", None]:
      path = INTERNAL_PATH_SEPARATOR  # root
      # convert it into absolute path
    path = os.path.normpath(path)
    if os.sep != INTERNAL_PATH_SEPARATOR:
      path = path.replace(os.sep, INTERNAL_PATH_SEPARATOR)

    # for this special case of anomaly when double // may appear
    path = path.replace(2 * INTERNAL_PATH_SEPARATOR, INTERNAL_PATH_SEPARATOR)

    return S_OK(path)

  #############################################################################
  def get(self, path="/"):
    """get a node for a given path."""
    path = self.getAbsolutePath(path)['Value']
    entity = self._getEntity(path)
    if not isinstance(entity, objects.Entity):
      gLogger.error(path + " doesn't exist!")
      # raise ValueError, "Invalid path %s" % path
    return S_OK(entity)

  #############################################################################
  def _getEntity(self, path):
    """This is not doing anything at the moment."""
    try:
      entity = self.__entityCache[path][0]
      gLogger.debug("getting " + str(path) + " from the cache")
      return entity
    except ValueError as ex:
      # not cached so far
      gLogger.debug(str(path) + " not in cache. Fetching..." + ex)

    # Second try

    try:
      gLogger.debug("getting " + str(path) + " eventually from the cache")
      entity = self.__entityCache[path][0]
      return entity
    except ValueError as ex:
      # still not in the cache... wrong path
      gLogger.warn(str(path) + " seems to be a wrong path" + ex)
      return None

    return entity

  #############################################################################
  @staticmethod
  def getNumberOfEvents(files):
    """statistics."""
    esum = 0
    for lfn in files:
      esum += int(lfn['EventStat'])
    return esum

  #############################################################################
  def getJobInfo(self, lfn):
    """job info."""
    result = self.db_.getJobInfo(lfn)
    value = None
    if result['OK']:
      dbResult = result['Value']
      for record in dbResult:
        value = {'DiracJobID': record[0], 'DiracVersion': record[1], 'EventInputStat': record[2],
                 'ExecTime': record[3], 'FirstEventNumber': record[4],
                 'Location': record[5], 'Name': record[6], 'NumberofEvents': record[7],
                 'StatisticsRequested': record[8], 'WNCPUPOWER': record[9],
                 'CPUTime': record[10], 'WNCACHE': record[11],
                 'WNMEMORY': record[12], 'WNMODEL': record[13],
                 'WORKERNODE': record[14], 'WNCPUHS06': record[15], 'TotalLuminosity': record[17]}
    else:
      gLogger.error(result['Message'])
    return value

  #############################################################################
  def getLimitedFiles(self, selectionDict, sortDict, startItem, maxitems):
    """web."""
    if self.parameter_ == self.__bookkeepingParameters[0]:
      return self._getLimitedFilesConfigParams(selectionDict, sortDict, startItem, maxitems)
    elif self.parameter_ == self.__bookkeepingParameters[1]:
      return self._getLimitedFilesEventTypeParams(selectionDict, sortDict, startItem, maxitems)
    elif self.parameter_ == self.__bookkeepingParameters[2]:
      return self._getLimitedFilesProductions(selectionDict, sortDict, startItem, maxitems)
    elif self.parameter_ == self.__bookkeepingParameters[3]:
      return self._getLimitedFilesRuns(selectionDict, sortDict, startItem, maxitems)

  #############################################################################
  def _getDataSetTree1(self, selectionDict):
    # it is the configname configversion(default) query.
    # The input parameter is a path and it constructs the dictionary.
    """input dictionary."""
    path = selectionDict['fullpath']
    levels, processedPath, procpass = self.getLevelAndPath(path)
    gLogger.debug(str(levels))
    retVal = procpass.split('/')[1:]
    for i in retVal:
      processedPath.remove(i)

    if self.advancedQuery_:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'ConditionDescription': processedPath[2],
                 'EventType': processedPath[3],
                 'Production': processedPath[4],
                 'FileType': processedPath[5]}
    else:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'ConditionDescription': processedPath[2],
                 'EventType': processedPath[3],
                 'FileType': processedPath[4]}
    in_dict['ProcessingPass'] = procpass
    in_dict['fullpath'] = path
    return in_dict

  #############################################################################
  def _getLimitedFilesConfigParams(self, selectionDict, sortDict, startItem, maxitems):
    """input dictionary."""
    selection = self._getDataSetTree1(selectionDict)
    return self.__getFiles(selection, sortDict, startItem, maxitems)

  #############################################################################
  def _getDataSetTree2(self, selectionDict):  # it is the event type based query
    """input dictionary of the event type tree."""
    path = selectionDict['fullpath']
    levels, processedPath, procpass = self.getLevelAndPath(path)
    gLogger.debug(str(levels))
    retVal = procpass.split('/')[1:]
    for i in retVal:
      processedPath.remove(i)

    if self.advancedQuery_:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'EventType': processedPath[2],
                 'ConditionDescription': processedPath[3],
                 'Production': processedPath[4],
                 'FileType': processedPath[5]}
    else:
      in_dict = {'ConfigName': processedPath[0],
                 'ConfigVersion': processedPath[1],
                 'EventType': processedPath[2],
                 'ConditionDescription': processedPath[3],
                 'FileType': processedPath[4]}
    in_dict['ProcessingPass'] = procpass
    in_dict['fullpath'] = path
    return in_dict

  #############################################################################
  def _getLimitedFilesEventTypeParams(self, selectionDict, sortDict, startItem, maxitems):
    """input dictionary."""
    selection = self._getDataSetTree2(selectionDict)
    return self.__getFiles(selection, sortDict, startItem, maxitems)

  #############################################################################
  def _getDataSetTree3(self, selectionDict):  # production based query
    """input dictionary."""
    path = selectionDict['fullpath']
    levels, processedPath, procpass = self.getLevelAndPath(path)
    gLogger.debug(str(levels))
    gLogger.debug(procpass)

    in_dict = {'Production': processedPath[0], 'EventType': processedPath[1], 'FileType': processedPath[2]}
    in_dict['fullpath'] = path
    return in_dict

  #############################################################################
  def _getLimitedFilesProductions(self, selectionDict, sortDict, startItem, maxitems):
    """input dictionary."""
    selection = self._getDataSetTree3(selectionDict)
    return self.__getFiles(selection, sortDict, startItem, maxitems)

  #############################################################################
  def _getDataSetTree4(self, selectionDict):  # run based query
    """input dictionary."""
    path = selectionDict['fullpath']
    levels, processedPath, procpass = self.getLevelAndPath(path)
    gLogger.debug(str(levels))
    retVal = procpass.split('/')[1:]
    for i in retVal:
      processedPath.remove(i)

    in_dict = {'RunNumber': processedPath[0],
               'EventType': processedPath[1],
               'FileType': processedPath[2],
               'ProcessingPass': procpass}
    in_dict['fullpath'] = path
    return in_dict

  #############################################################################
  def _getLimitedFilesRuns(self, selectionDict, sortDict, startItem, maxitems):
    """input dictionary."""
    selection = self._getDataSetTree4(selectionDict)
    return self.__getFiles(selection, sortDict, startItem, maxitems)

  #############################################################################
  def __getFiles(self, in_dict, sortDict, startItem, maxitems):
    """returns the files."""
    totalrecords = 0
    nbOfEvents = 0
    filesSize = 0
    lumi = 0
    selection = in_dict
    in_dict['Visible'] = 'Y'
    if len(sortDict) > 0:
      in_dict['DataQuality'] = self.__getSelectedQualities()
      res = self.db_.getFilesSummary(in_dict)
      if not res['OK']:
        gLogger.error(res['Message'])
      else:
        records = res['Value']['Records']
        totalrecords = records[0][0]
        nbOfEvents = records[0][1]
        filesSize = records[0][2]
        lumi = records[0][3]
    records = []
    parametersNames = []
    if startItem > -1 and maxitems != 0:
      in_dict['StartItem'] = startItem
      in_dict['MaxItem'] = maxitems
      in_dict['DataQuality'] = self.__getSelectedQualities()
      result = self.db_.getLimitedFiles(in_dict)

      if result['OK']:
        parametersNames = result['Value']['ParameterNames']
        records = result['Value']['Records']
        if totalrecords == 0:
          totalrecords = result['Value']['TotalRecords']
      else:
        gLogger.error(result['Message'])
        return result

    return S_OK({'TotalRecords': totalrecords,
                 'ParameterNames': parametersNames,
                 'Records': records,
                 'Extras': {'Selection': selection,
                            'GlobalStatistics': {'Number of Events': nbOfEvents,
                                                 'Files Size': filesSize,
                                                 'Luminosity': lumi}}})

  #############################################################################
  def getAncestors(self, files, depth):
    """ancestors of given files."""
    return self.db_.getFileAncestors(files, depth)

  #############################################################################
  def getLogfile(self, filename):
    """log file."""
    return self.db_.getFileCreationLog(filename)

  #############################################################################
  def writePythonOrJobOptions(self, startItem, maxitems, path, savetype):
    """create Gaudi Card."""
    result = None
    dataset = None
    if self.parameter_ == self.__bookkeepingParameters[0]:
      result = self._getLimitedFilesConfigParams({'fullpath': path}, {}, startItem, maxitems)
      dataset = self._getDataSetTree1({'fullpath': path})
    elif self.parameter_ == self.__bookkeepingParameters[1]:
      result = self._getLimitedFilesEventTypeParams({'fullpath': path}, {}, startItem, maxitems)
      dataset = self._getDataSetTree2({'fullpath': path})
    elif self.parameter_ == self.__bookkeepingParameters[2]:
      result = self._getLimitedFilesProductions({'fullpath': path}, {}, startItem, maxitems)
      dataset = self._getDataSetTree3({'fullpath': path})
    elif self.parameter_ == self.__bookkeepingParameters[3]:
      result = self._getLimitedFilesRuns({'fullpath': path}, {}, startItem, maxitems)
      dataset = self._getDataSetTree4({'fullpath': path})

    if 'TotalRecords' in result['Value'] and result['Value']['TotalRecords'] > 0:
      records = result['Value']['Records']
      params = result['Value']['ParameterNames']
      # The list has to be converted to dictionary
      nameIndex = params.index('Name')
      files = {}
      for rec in records:
        lfn = rec[nameIndex]
        files[lfn] = dict(zip(params, rec))
      return self.writeJobOptions(files, optionsFile='',
                                  savedType=savetype, catalog=None,
                                  savePfn=None, dataset=dataset)
    else:
      return S_ERROR("Error discovered during the option file creation!")

  #############################################################################
  def getLimitedInformations(self, startItem, maxitems, path):
    """statistics."""
    result = None
    if self.parameter_ == self.__bookkeepingParameters[0]:
      result = self._getLimitedFilesConfigParams({'fullpath': path}, {'need': 0}, startItem, maxitems)
    elif self.parameter_ == self.__bookkeepingParameters[1]:
      result = self._getLimitedFilesEventTypeParams({'fullpath': path}, {'need': 0}, startItem, maxitems)
    elif self.parameter_ == self.__bookkeepingParameters[2]:
      result = self._getLimitedFilesProductions({'fullpath': path}, {'need': 0}, startItem, maxitems)
    elif self.parameter_ == self.__bookkeepingParameters[3]:
      result = self._getLimitedFilesRuns({'fullpath': path}, {'need': 0}, startItem, maxitems)

    if 'TotalRecords' in result['Value'] and result['Value']['TotalRecords'] > 0:
      nbe = 0
      fsize = 0
      nbfiles = 0
      params = result['Value']['ParameterNames']
      for lfn in result['Value']['Records']:
        nbfiles += 1
        if lfn[params.index('EventStat')] is not None:
          nbe += int(lfn[params.index('EventStat')])
        if lfn[params.index('FileSize')] is not None:
          fsize += int(lfn[params.index('FileSize')])
      return S_OK({'Number of Events': nbe, 'Files Size': fsize, 'Number of files': nbfiles})
    else:
      return S_ERROR("Error discoverd during the option file creation!")

  #############################################################################
  def writeJobOptions(self, files, optionsFile=None, savedType=None, catalog=None, savePfn=None, dataset=None):
    """create options file.

    :params dict files: LFNs for which to write the options
    :params str optionsFile: file name where options might be written (iff given)
    :params str savedType: if savedType == 'txt', return the list of LFNs
    :params str catalog: add catalog or not
    :params dict savePfn: file type versions
    :params dict dataset: metadata about the dataset

    :returns: str with options
    """
    if savedType == 'py' or savedType is None:
      # get list of event types
      evtTypes = self.__createEventTypeList(files)

      string = self.__addGaudiheader(evtTypes)
      string += self.__addDatasetCreationMetadata(dataset)

      filesandformats = self.__getFilesandFormats(savePfn, files)
      string += self.__createFormatString(filesandformats)

      if catalog:
        string += "\nFileCatalog().Catalogs += [ 'xmlcatalog_file:" + os.path.basename(catalog) + "' ]\n"
    elif savedType == 'txt':
      # Only return the list of LFNs
      string = '\n'.join(str(lfn) for lfn in files) + '\n'
    else:
      raise NotImplementedError(savedType)

    if optionsFile:
      # Write options file if requested
      with open(optionsFile, 'wt') as fd:
        fd.write(string)

    # Always return the string
    return string

  #############################################################################
  def __addDatasetCreationMetadata(self, dataset):
    """it adds the metadata information about the dataset creation."""
    string = ''
    if dataset:
      string += "\n%s Extra information about the data processing phases:\n" % (self.comment)
      retVal = self.db_.getStepsMetadata(dataset)
      if retVal['OK']:
        for ppass, record in retVal['Value']['Records'].items():
          ppass = dataset.get('ProcessingPass', ppass)
          string += "\n%s Processing Pass: '%s' \n\n" % (self.comment, ppass)
          for i in record:
            string += "%s %s : %s \n" % (self.comment, i[0], i[1])
    return string

  #############################################################################
  @staticmethod
  def __createFormatString(filesandformats):
    """It generates the Root format option file."""
    string = "\nfrom Gaudi.Configuration import * "
    string += "\nfrom GaudiConf import IOHelper\n"
    for fileFormat, lfns in filesandformats.items():
      if fileFormat:
        string += "IOHelper('%s').inputFiles([\n" % fileFormat
      else:
        string += "IOHelper().inputFiles([\n"
      string += '\n'.join("'LFN:%s'," % lfn for lfn in lfns)
      string += '\n], clear=True)\n'

    return string

  #############################################################################
  def __addGaudiheader(self, evtTypes):
    """it creates the header of the job option."""
    string = self.comment + "GAUDI jobOptions generated on " + time.asctime() + "\n"
    if evtTypes:
      string += self.comment + "Contains event types : \n"
    for evtType in sorted(evtTypes):
      string += self.comment + "  %8d - %d files - %d events - %.2f GBytes\n" % (evtType,
                                                                                 evtTypes[evtType][0],
                                                                                 evtTypes[evtType][1],
                                                                                 evtTypes[evtType][2])
    return string

  #############################################################################
  def __getFilesandFormats(self, savePfn, files):
    """It returns a list of lfns for each file format (as a dictionary)"""
    filesandformats = {}
    if savePfn:
      # we have to decide the file type version.
      # This variable contains the file type version, if it is empty I check in the bkk
      lfns = []
      for lfn in files:
        if lfn not in savePfn:
          lfns.append(lfn)
          continue
        if isinstance(savePfn[lfn], list):
          fileFormat = savePfn[lfn][0]['pfntype']
        else:
          fileFormat = savePfn[lfn]['pfntype']
        filesandformats.setdefault(fileFormat, []).append(lfn)

    else:
      lfns = list(files)

    if lfns:
      # Get file type version from BK
      retVal = self.db_.getFileTypeVersion(lfns)
      if retVal['OK']:
        for lfn, fileFormat in retVal['Value'].items():
          filesandformats.setdefault(fileFormat, []).append(lfn)
          lfns.remove(lfn)
      # If no persistency is found, set it to None
      if lfns:
        filesandformats[None] = lfns
    return filesandformats

  #############################################################################
  @staticmethod
  def __createEventTypeList(files):
    """It creates a dictionary which contains the event types and the size of
    the data set."""
    evtTypes = {}
    if not isinstance(files, dict):
      return evtTypes
    for metadata in files.values():
      evtType = metadata.get('EventType')
      if evtType:
        try:
          stat = int(metadata['EventStat'])
        except (KeyError, ValueError, TypeError):
          stat = 0

        info = evtTypes.setdefault(int(evtType), [0, 0, 0.])
        info[0] += 1
        info[1] += stat
        size = metadata.get('FileSize')
        if size is not None:
          info[2] += int(size) / 1000000000.
    return evtTypes

  #############################################################################
  def getProcessingPassSteps(self, in_dict):
    """steps for a given processing pass."""
    return self.db_.getProcessingPassSteps(in_dict)

  #############################################################################
  def getMoreProductionInformations(self, prodid):
    """production statistics."""
    return self.db_.getMoreProductionInformations(prodid)

  #############################################################################
  def getAvailableProductions(self):
    """available productions."""
    return self.db_.getAvailableProductions()

  #############################################################################
  def getFileHistory(self, lfn):
    """history."""
    return self.db_.getFileHistory(lfn)

  #############################################################################
  def getProductionProcessingPassSteps(self, in_dict):
    """step of a production."""
    return self.db_.getProductionProcessingPassSteps(in_dict)

  #############################################################################
  def getAvailableDataQuality(self):
    """all the existing data qualities."""
    return self.db_.getAvailableDataQuality()

  #############################################################################
  def setDataQualities(self, values):
    """setting data quality."""
    self.dataQualities_ = values

  def __getSelectedQualities(self):
    """data quality."""
    return [flag for flag, val in self.dataQualities_.items() if val is True]

  #############################################################################
  def getStepsMetadata(self, bkDict):
    """It is a wrapper to the bookkeeping client."""
    return self.db_.getStepsMetadata(bkDict)

  #############################################################################
  def getFilesWithMetadata(self, dataset):
    """it sets the file types.

    :param dict dataset: it is a bookkeeping dictionary, which contains the conditions used to retreive the lfns
    :return: S_OK lfns with metadata
    """
    return self.db_.getFilesWithMetadata(dataset)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.AccountingSystem.private.Plotters.DataStoragePlotter.

DataStoragePlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter.BaseReporter
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

from LHCbDIRAC.AccountingSystem.Client.Types.DataStorage import DataStorage

__RCSID__ = "$Id$"

# FIXME: refactor _reportMethods
# FIXME: refactor _plotMethods


class DataStoragePlotter(BaseReporter):
  """DataStoragePlotter as extension of BaseReporter."""

  _typeName = "DataStorage"
  _typeKeyFields = [dF[0] for dF in DataStorage().definitionKeyFields]

  # Catalog Space

  _reportCatalogSpaceName = "LFN size"

  def _reportCatalogSpace(self, reportRequest):
    """Reports about the LFN size and the catalog space from the accounting.

    :param reportRequest: <dict>
      { 'grouping'       : 'EventType',
        'groupingFields' : ( '%s', [ 'EventType' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'EventType' : 'Full stream' }
      }

    returns S_OK / S_ERROR
      { 'graphDataDict': { 'Full stream': { 1355616000L: 935.38852424691629,
                                            1355702400L: 843.84448707482204 }
                         },
        'data'         : { 'Full stream': { 1355616000L: 935388.52424691629,
                                            1355702400L: 843844.48707482207 }
                         },
        'unit'         : 'GB',
        'granularity'  : 86400
      }
    """

    if reportRequest['grouping'] == "StorageElement":
      return S_ERROR("Grouping by storage element when requesting lfn info makes no sense")

    selectString = self._getSelectStringForGrouping(reportRequest['groupingFields'])

    selectFields = (selectString + ", %s, %s, SUM(%s/%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'LogicalSize', 'entriesInBucket']
                    )

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxValue = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "bytes")

    # 3rd variable unused ( maxValue )
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def _plotCatalogSpace(self, reportRequest, plotInfo, filename):
    """Creates <filename>.png file containing information regarding the LFN
    size and the catalog space.

    :param reportRequest: <dict>
       { 'grouping'       : 'EventType',
         'groupingFields' : ( '%s', [ 'EventType' ] ),
         'startTime'      : 1355663249.0,
         'endTime'        : 1355749690.0,
         'condDict'       : { 'EventType' : 'Full stream' }
       }
    :param plotInfo: <dict> ( output of _reportCatalogSpace )
       { 'graphDataDict' : { 'Full stream' : { 1355616000L: 4.9003546130956819,
                                               1355702400L: 4.9050379437065059 }
                           },
         'data'          : { 'Full stream' : { 1355616000L: 4900354613.0956821,
                                               1355702400L: 4905037943.7065058 }
                           },
         'unit'          : 'PB',
         'granularity'   : 86400
        }
    :param filename: <str>
      '_plotCatalogSpace'

    returns S_OK / S_ERROR
       { 'plot': True, 'thumbnail': False }
    """

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']
    span = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {'title': "LFN space usage grouped by %s" % reportRequest['grouping'],
                'starttime': startTime,
                'endtime': endTime,
                'span': span,
                'ylabel': plotInfo['unit']}

    dataDict = self._fillWithZero(span, startTime, endTime, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # .............................................................................
  # Catalog Files

  _reportCatalogFilesName = "LFN files"

  def _reportCatalogFiles(self, reportRequest):
    """Reports about the LFN files and the catalog files from the accounting.

    :param reportRequest: <dict>
      { 'grouping'       : 'EventType',
        'groupingFields' : ( '%s', [ 'EventType' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'EventType' : 'Full stream' }
      }

    returns S_OK / S_ERROR
      { 'graphDataDict' : { 'Full stream' : { 1355616000L : 420.47885754501203,
                                              1355702400L : 380.35170637810842 }
                                            },
        'data'          : { 'Full stream' : { 1355616000L : 420.47885754501203,
                                              1355702400L : 380.35170637810842 }
                                            },
        'unit'          : 'files',
        'granularity'   : 86400
      }
    """

    if reportRequest['grouping'] == "StorageElement":
      return S_ERROR("Grouping by storage element when requesting lfn info makes no sense")

    selectString = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectString + ", %s, %s, SUM(%s/%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'LogicalFiles', 'entriesInBucket'
                                                          ]
                    )
    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxValue = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "files")

    # 3rd variable unused ( maxValue )
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def _plotCatalogFiles(self, reportRequest, plotInfo, filename):
    """Creates <filename>.png file containing information regarding the LFN
    files and the catalog files.

    :param reportRequest: <dict>
      { 'grouping'       : 'EventType',
        'groupingFields' : ( '%s', [ 'EventType' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'EventType' : 'Full stream' }
      }
    :param plotInfo: <dict> ( output of _reportCatalogFiles )
      { 'graphDataDict' : { 'Full stream' : { 1355616000L : 420.47885754501203,
                                              1355702400L : 380.35170637810842 }
                          },
        'data'          : { 'Full stream' : { 1355616000L : 420.47885754501203,
                                              1355702400L : 380.35170637810842 }
                          },
        'unit'          : 'files',
        'granularity'   : 86400
      }
    :param filename: <str>
      '_plotCatalogFiles'

    returns S_OK / S_ERROR
       { 'plot': True, 'thumbnail': False }
    """

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']
    span = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {'title': "Number of LFNs by %s" % reportRequest['grouping'],
                'starttime': startTime,
                'endtime': endTime,
                'span': span,
                'ylabel': plotInfo['unit']}

    dataDict = self._fillWithZero(span, startTime, endTime, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # .............................................................................
  # Physical Space

  _reportPhysicalSpaceName = "PFN size"

  def _reportPhysicalSpace(self, reportRequest):
    """Reports about the PFN size and the physical space from the accounting.

    :param reportRequest: <dict>
      { 'grouping'       : 'EventType',
        'groupingFields' : ( '%s', [ 'EventType' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'EventType' : 'Full stream' }
      }

    returns S_OK / S_ERROR
      { 'graphDataDict' : { 'Full stream' : { 1355616000L : 14.754501202,
                                              1355702400L : 15.237810842 }
                          },
        'data'          : { 'Full stream' : { 1355616000L : 14.754501202,
                                              1355702400L : 15.237810842 }
                          },
        'unit'          : 'MB',
        'granularity'   : 86400
      }
    """

    selectString = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectString + ", %s, %s, SUM(%s/%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'PhysicalSize', 'entriesInBucket'
                                                          ]
                    )
    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxValue = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "bytes")

    # 3rd variable unused ( maxValue )
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({
        'data': baseDataDict,
        'graphDataDict': graphDataDict,
        'granularity': granularity,
        'unit': unitName
    })

  def _plotPhysicalSpace(self, reportRequest, plotInfo, filename):
    """Creates <filename>.png file containing information regarding the PFN
    size and the physical space.

    :param reportRequest: <dict>
      { 'grouping'       : 'EventType',
        'groupingFields' : ( '%s', [ 'EventType' ] ),
        'startTime'      : 1355663249.0,
        'endTime'        : 1355749690.0,
        'condDict'       : { 'EventType' : 'Full stream' }
      }
    :param plotInfo: <dict> ( output of _reportPhysicalSpace )
      { 'graphDataDict' : { 'Full stream' : { 1355616000L : 14.754501202,
                                              1355702400L : 15.237810842 }
                          },
        'data'          : { 'Full stream' : { 1355616000L : 14.754501202,
                                              1355702400L : 15.237810842 }
                          },
        'unit'          : 'MB',
        'granularity'   : 86400
      }
    :param filename: <str>
      '_plotPhysicalSpace'

    returns S_OK / S_ERROR
       { 'plot': True, 'thumbnail': False }
    """

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']
    span = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {'title': "PFN space usage by %s" % reportRequest['grouping'],
                'starttime': startTime,
                'endtime': endTime,
                'span': span,
                'ylabel': plotInfo['unit']}

    dataDict = self._fillWithZero(span, startTime, endTime, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # .............................................................................
  # Physical Files

  _reportPhysicalFilesName = "PFN files"

  def _reportPhysicalFiles(self, reportRequest):
    """Reports about the PFN files and the physical files from the accounting.

    :param reportRequest: <dict>
       { 'grouping'       : 'EventType',
         'groupingFields' : ( '%s', [ 'EventType' ] ),
         'startTime'      : 1355663249.0,
         'endTime'        : 1355749690.0,
         'condDict'       : { 'EventType' : 'Full stream' }
       }

    returns S_OK / S_ERROR
      { 'graphDataDict' : { 'Full stream' : { 1355616000L : 42.47885754501203,
                                              1355702400L : 38.35170637810842 }
                          },
        'data'          : { 'Full stream' : { 1355616000L : 42.47885754501203,
                                              1355702400L : 38.35170637810842 }
                          },
        'unit'          : 'files',
        'granularity'   : 86400
      }
    """

    selectString = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectString + ", %s, %s, SUM(%s/%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          'PhysicalFiles', 'entriesInBucket'
                                                          ]
                    )
    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal
    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    accumMaxValue = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "files")

    # 3rd variable unused ( maxValue )
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def _plotPhysicalFiles(self, reportRequest, plotInfo, filename):
    """Creates <filename>.png file containing information regarding the PFN
    files and the physical files.

    :param reportRequest: <dict>
       { 'grouping'       : 'EventType',
         'groupingFields' : ( '%s', [ 'EventType' ] ),
         'startTime'      : 1355663249.0,
         'endTime'        : 1355749690.0,
         'condDict'       : { 'EventType' : 'Full stream' }
       }
    :param plotInfo: <dict> ( output of _reportPhysicalFiles )
      { 'graphDataDict' : { 'Full stream' : { 1355616000L: 4.9003546130956819,
                                              1355702400L: 4.9050379437065059 }
                          },
        'data'          : { 'Full stream' : { 1355616000L: 4900354613.0956821,
                                              1355702400L: 4905037943.7065058 }
                          },
        'unit'          : 'PB',
        'granularity'   : 86400
      }
    :param filename: <str>
      '_plotPhysicalFiles'

    return S_OK / S_ERROR
       { 'plot': True, 'thumbnail': False }
    """

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']
    span = plotInfo['granularity']
    dataDict = plotInfo['graphDataDict']

    metadata = {'title': "Number of PFNs by %s" % reportRequest['grouping'],
                'starttime': startTime,
                'endtime': endTime,
                'span': span,
                'ylabel': plotInfo['unit']}

    dataDict = self._fillWithZero(span, startTime, endTime, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""In a transformation, flush a list of runs or runs that are flushed in the
transformation used in BKQuery."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC import gLogger
  from DIRAC.Core.Base import Script

  Script.registerSwitch('', 'Runs=', '   list of runs to flush (comma separated, ranges r1:r2)')
  Script.registerSwitch('', 'NoAction', '   No action taken, just give stats')
  Script.registerSwitch('', 'Active', '   If used, selects all active runs')
  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] <TransID>[,<TransID2>...]' % Script.scriptName, ]))
  Script.parseCommandLine()

  args = Script.getPositionalArgs()
  action = True
  runList = []
  active = False
  for switch, val in Script.getUnprocessedSwitches():
    if switch == 'NoAction':
      action = False
    elif switch == 'Active':
      active = True
    elif switch == 'Runs':
      try:
        runs = val.split(',')
        for run in runs:
          if run.isdigit():
            runList.append(int(run))
          else:
            runRange = run.split(':')
            if len(runRange) == 2 and runRange[0].isdigit() and runRange[1].isdigit():
              runList += range(int(runRange[0]), int(runRange[1]) + 1)
      except Exception as x:
        gLogger.exception('Bad run parameter', lException=x)

  if len(args) != 1:
    gLogger.fatal("Specify transformation number...")
    Script.showHelp(exitCode=1)
  else:
    ids = args[0].split(",")
    idList = []
    for id in ids:
      r = id.split(':')
      if len(r) > 1:
        for i in range(int(r[0]), int(r[1]) + 1):
          idList.append(i)
      else:
        idList.append(int(r[0]))

  from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  transClient = TransformationClient()

  for transID in idList:
    res = transClient.getTransformationRuns({'TransformationID': transID})
    if not res['OK']:
      gLogger.fatal("Error getting runs for transformation %s" % transID, res['Message'])
      DIRAC.exit(1)
    runs = res['Value']
    runs.sort(cmp=(lambda r1, r2: int(r1['RunNumber'] - r2['RunNumber'])))

    if not runList:
      if active:
        # Flush all Active runs in that transformation
        toBeFlushed = [run['RunNumber'] for run in runs if run['Status'] != 'Flush']
      else:
        # Get transformationID used in the BK query, to find out which runs are flushed in there
        res = transClient.getBookkeepingQuery(transID)
        if not res['OK']:
          gLogger.fatal("Error getting BK query for transformation %s" % transID, res['Message'])
          DIRAC.exit(1)
        queryProd = res['Value'].get('ProductionID')
        if not queryProd:
          gLogger.fatal("Transformation is not based on another one")
          DIRAC.exit(0)

        res = transClient.getTransformationRuns({'TransformationID': queryProd})
        if not res['OK']:
          gLogger.fatal("Error getting runs for transformation %s" % queryProd, res['Message'])
          DIRAC.exit(1)
        queryRuns = res['Value']
        # Get the list of runs flushed in the parent production
        flushedRuns = [run['RunNumber'] for run in queryRuns if run['Status'] == 'Flush']

        toBeFlushed = []
        # Get runs not flushed in current production but flushed in the parent production
        for run in [run['RunNumber'] for run in runs if run['Status'] != 'Flush' and run['RunNumber'] in flushedRuns]:
          missing = -1
          res = transClient.getTransformationFiles({'TransformationID': queryProd, 'RunNumber': run})
          if not res['OK']:
            gLogger.fatal("Error getting files for run %d" % run, res['Message'])
          else:
            runFiles = res['Value']
            missing = 0
            for rFile in runFiles:
              if rFile['Status'] in ('Unused', 'Assigned', 'MaxReset'):
                missing += 1
            if not missing:
              toBeFlushed.append(run)

    else:
      toBeFlushed = sorted(set(runList) & set([run['RunNumber'] for run in runs]))

    ok = 0
    gLogger.always("Runs %s flushed in transformation %s:" %
                   ('being' if action else 'to be', transID), ','.join([str(run) for run in toBeFlushed]))
    for run in toBeFlushed:
      res = transClient.setTransformationRunStatus(transID, run, 'Flush') if action else {'OK': True}
      if not res['OK']:
        gLogger.fatal("Error setting run %s to Flush in transformation %s" % (run, transID), res['Message'])
      else:
        ok += 1
    gLogger.always('%d runs set to Flush in transformation %d' % (ok, transID))


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module to upload specified job output files according to the parameters
defined in the user workflow."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Utilities import DEncode
from DIRAC.RequestManagementSystem.Client.Request import Operation
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.DataManagementSystem.Client.DataUsageClient import DataUsageClient


class FileUsage(ModuleBase):

  #############################################################################

  def __init__(self, bkClient=None, dm=None):
    """Module initialization."""
    self.log = gLogger.getSubLogger("FileUsage")
    super(FileUsage, self).__init__(self.log, bkClientIn=bkClient, dm=dm)
    self.version = __RCSID__
    self.dataUsageClient = DataUsageClient()

  #############################################################################

  def _resolveInputVariables(self):
    """By convention the module parameters are resolved here."""
    super(FileUsage, self)._resolveInputVariables()

    dirDict = {}
    if self.inputDataList:
      for inputFile in self.inputDataList:
        strippedDir = os.path.join(os.path.dirname(inputFile), '')
        if not strippedDir:
          self.log.error('File specified without directory! ', inputFile)
        else:
          dirDict[strippedDir] = dirDict.setdefault(strippedDir, 0) + 1
      self.log.info('Popularity report = ', dirDict)
    else:
      self.log.info('No input data specified for this job')

    return S_OK(dirDict)

  #############################################################################

  def execute(self, production_id=None, prod_job_id=None, wms_job_id=None,
              workflowStatus=None, stepStatus=None,
              wf_commons=None, step_commons=None,
              step_id=None, step_number=None):
    """Main execution function."""

    try:

      super(FileUsage, self).execute(self.version, production_id, prod_job_id, wms_job_id,
                                     workflowStatus, stepStatus,
                                     wf_commons, step_commons, step_number, step_id)

      result = self._resolveInputVariables()
      if not result['OK']:
        self.log.error(result['Message'])
        return S_OK()
      dirDict = result['Value']

      self.request.RequestName = 'job_%d_request.xml' % self.jobID
      self.request.JobID = self.jobID
      self.request.SourceComponent = "Job_%d" % self.jobID

      if dirDict:
        result = self._reportFileUsage(dirDict)
        if not result['OK']:
          self.log.error(result['Message'])
          return S_OK()
        self.log.info("Reporting input file usage successful!")
      else:
        self.log.info("No input data usage to report!")

      return S_OK('File Usage reported successfully')

    except Exception as e:  # pylint:disable=broad-except
      self.log.exception("Failure in FileUsage execute module", lException=e)
      return S_ERROR(str(e))

    finally:
      super(FileUsage, self).finalize(self.version)

  #############################################################################

  def _reportFileUsage(self, dirDict):
    """Send the data usage report (site,dirDict) where dirDict = {'Dataset':NumberOfHits}
    example: {'/lhcb/certification/test/ALLSTREAMS.DST/00000002/0000/': 1,
    '/lhcb/LHCb/Collision11/BHADRON.DST/00012957/0000/': 2}
    """
    self.log.verbose('Reporting input file usage:')
    for entry in dirDict:
      self.log.verbose('%s:%s' % (entry, dirDict[entry]))
    localSite = gConfig.getValue('/LocalSite/Site', 'UNKNOWN')
    try:
      localSite = localSite.split('.')[1]
    except BaseException:
      pass
    self.log.verbose('Using Site Name: %s' % (localSite))

    if self._enableModule():
      usageStatus = self.dataUsageClient.sendDataUsageReport(localSite, dirDict)
      if not usageStatus['OK']:
        self.log.error('Could not send data usage report, preparing a DISET failover request object')
        self.log.verbose(usageStatus['rpcStub'])
        forwardDISETOp = Operation()
        forwardDISETOp.Type = "ForwardDISET"
        forwardDISETOp.Arguments = DEncode.encode(usageStatus['rpcStub'])
        self.request.addOperation(forwardDISETOp)
        self.workflow_commons['Request'] = self.request
    else:
      self.log.info('Would have attempted to report %s at %s' % (dirDict, localSite))
      return S_OK()

    return S_OK()

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module holding ProductionRequestClient class."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Base.Client import Client, createClient

__RCSID__ = "$Id$"


@createClient('ProductionManagement/ProductionRequest')
class ProductionRequestClient(Client):
  """This class expose the methods of the Production Request Service."""

  def __init__(self, url=None, **kwargs):
    """c'tor.

    :param str url: can specify a specific URL
    """
    super(ProductionRequestClient, self).__init__(**kwargs)
    self.setServer('ProductionManagement/ProductionRequest')
    if url:
      self.setServer(url)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


def usage():
  from DIRAC.Core.Base import Script
  print('Usage: %s <Production ID> <DIRAC Status>' % (Script.scriptName))
  DIRAC.exit(2)


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script
  Script.parseCommandLine(ignoreErrors=True)
  from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

  args = Script.getPositionalArgs()

  if len(args) < 2:
    usage()

  diracProd = DiracProduction()
  prodID = args[0]
  stat = args[1]
  result = diracProd.getProductionApplicationSummary(prodID, status=stat, printOutput=True)
  if result['OK']:
    DIRAC.exit(0)
  elif 'Message' in result:
    print('Getting production application status summary failed with message:\n%s' % (result['Message']))
    DIRAC.exit(2)
  else:
    print('Null result for getProductionApplicationSummary() call')
    DIRAC.exit(2)


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""dirac-production-runjoblocal.

Module created to run failed jobs locally on a CVMFS-configured machine.
It creates the necessary environment, downloads the necessary files, modifies the necessary
files and runs the job

Usage:
  dirac-production-runjoblocal (job ID) (Data imput mode) -  No parenthesis
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import sys
import os
import shutil
import ssl

if sys.version_info < (3,):
  from urllib2 import urlopen as url_library_urlopen  # pylint: disable=no-name-in-module,import-error
  from urllib2 import URLError as url_library_URLError  # pylint: disable=no-name-in-module,import-error
else:
  from urllib.request import urlopen as url_library_urlopen  # pylint: disable=no-name-in-module,import-error
  from urllib.error import URLError as url_library_URLError  # pylint: disable=no-name-in-module,import-error

from DIRAC import S_OK
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


def __runSystemDefaults(jobID=None):
  """Creates the environment for running the job and returns the path for the
  other functions."""
  tempdir = "LHCbjob" + str(jobID) + "temp"
  os.environ['VO_LHCB_SW_DIR'] = "/cvmfs/lhcb.cern.ch"
  mkDir(tempdir)

  basepath = os.getcwd()
  return basepath + os.path.sep + tempdir + os.path.sep


def __downloadJobDescriptionXML(jobID, basepath):
  """Downloads the jobDescription.xml file into the temporary directory
  created."""
  from DIRAC.Interfaces.API.Dirac import Dirac
  jdXML = Dirac()
  jdXML.getInputSandbox(jobID, basepath)


def __modifyJobDescription(jobID, basepath, downloadinputdata):
  """Modifies the jobDescription.xml to, instead of DownloadInputData, it uses
  InputDataByProtocol."""
  if not downloadinputdata:
    from xml.etree import ElementTree as et
    archive = et.parse(basepath + "InputSandbox" + str(jobID) + os.path.sep + "jobDescription.xml")
    for element in archive.iter():
      if element.text == "DIRAC.WorkloadManagementSystem.Client.DownloadInputData":
        element.text = "DIRAC.WorkloadManagementSystem.Client.InputDataByProtocol"
        archive.write(basepath + "InputSandbox" + str(jobID) + os.path.sep + "jobDescription.xml")
        return S_OK("Job parameter changed from DownloadInputData to InputDataByProtocol.")


def __downloadPilotScripts(basepath):
  """
  Downloads the scripts necessary to configure the pilot
  """
  context = ssl._create_unverified_context()
  for fileName in ['dirac-pilot.py', 'dirac-install.py',
                   'pilotCommands.py', 'pilotTools',
                   'MessageSender', 'PilotLogger.py', 'PilotLoggerTools.py']:
    remoteFile = url_library_urlopen(
        os.path.join('https://raw.githubusercontent.com/DIRACGrid/Pilot/master/Pilot/', fileName),
        timeout=10,
        context=context)
    with open(fileName, 'wb') as localFile:
      localFile.write(remoteFile.read())

  remoteFile = url_library_urlopen(
      os.path.join('https://gitlab.cern.ch/lhcb-dirac/LHCbPilot/-/raw/master/LHCbPilot/LHCbPilotCommands.py'),
      timeout=10,
      context=context)
  with open('LHCbPilotCommands.py', 'wb') as localFile:
    localFile.write(remoteFile.read())
  remoteFile = url_library_urlopen(
      os.path.join('https://gitlab.cern.ch/lhcb-dirac/LHCbPilot/-/raw/master/LHCbPilot/LHCbPilotTools.py'),
      timeout=10,
      context=context)
  with open('LHCbPilotTools.py', 'wb') as localFile:
    localFile.write(remoteFile.read())


def __configurePilot(basepath):
  """Configures the pilot."""
  pilotCmd = "dirac-pilot.py -S LHCb-Production -l LHCb "
  pilotCmd += "-C dips://lhcb-conf-dirac.cern.ch:9135/Configuration/Server "
  pilotCmd += "-N ce.debug.ch -Q default -n DIRAC.JobDebugger.cern -M 1 "
  pilotCmd += "-E LHCbPilot "
  pilotCmd += "-X LHCbConfigureBasics,LHCbConfigureSite,LHCbConfigureArchitecture,LHCbConfigureCPURequirements -dd"
  out = os.system("python " + basepath + pilotCmd)
  if not out:
    directory = os.path.expanduser('~') + os.path.sep
    os.rename(directory + '.dirac.cfg', directory + '.dirac.cfg.old')
    shutil.copyfile(directory + 'pilot.cfg', directory + '.dirac.cfg')
    return S_OK("Pilot successfully configured.")

#   else:
#     some DErrno message


def __runJobLocally(jobID, basepath):
  """Runs the job!"""
  from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob
  localJob = LHCbJob(basepath + "InputSandbox" + str(jobID) + os.path.sep + "jobDescription.xml")
  localJob.setInputSandbox(os.getcwd() + "pilot.cfg")
  localJob.setConfigArgs(os.getcwd() + "pilot.cfg")
  os.chdir(basepath)
  localJob.runLocal()


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.registerSwitch('D:', 'Download=', 'Defines data acquisition as DownloadInputData')
  Script.registerSwitch('P:', 'Protocol=', 'Defines data acquisition as InputDataByProtocol')
  Script.parseCommandLine(ignoreErrors=False)

  Script.setUsageMessage(__doc__ + '\n'.join([
      '\nUsage:',
      'dirac-production-runjoblocal [Data imput mode] [job ID]'
      '\nArguments:',
      '  Download (Job ID): Defines data acquisition as DownloadInputData',
      '  Protocol (Job ID): Defines data acquisition as InputDataByProtocol\n']))

  _downloadinputdata = False
  _jobID = None

  for switch in Script.getUnprocessedSwitches():
    if switch[0] in ('D', 'Download'):
      _downloadinputdata = True
      _jobID = switch[1]
    if switch[0] in ('P', 'Protocol'):
      _downloadinputdata = False
      _jobID = switch[1]

  usedDir = os.path.expanduser('~') + os.path.sep
  try:
    _path = __runSystemDefaults(_jobID)

    __downloadJobDescriptionXML(_jobID, _path)

    __modifyJobDescription(_jobID, _path, _downloadinputdata)

    __downloadPilotScripts(_path)

    __configurePilot(_path)

    __runJobLocally(_jobID, _path)

  finally:
    os.chdir(usedDir)
    os.rename(usedDir + '.dirac.cfg.old', usedDir + '.dirac.cfg')


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Add files to a transformation."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.TransformationSystem.Utilities.ScriptUtilities import getTransformations

  import os
  import DIRAC
  from DIRAC import gLogger
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script

  dmScript = DMScript()
  dmScript.registerFileSwitches()

  Script.registerSwitch('', 'NoRunInfo', '   Use if no run information is required')
  Script.registerSwitch("", "Chown=", "   Give user/group for chown of the directories of files in the FC")

  Script.parseCommandLine(ignoreErrors=True)

  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] ...' % Script.scriptName, ]))

  runInfo = True
  userGroup = None

  switches = Script.getUnprocessedSwitches()
  for opt, val in switches:
    if opt == 'NoRunInfo':
      runInfo = False
    elif opt == 'Chown':
      userGroup = val.split('/')
      if len(userGroup) != 2 or not userGroup[1].startswith('lhcb_'):
        gLogger.fatal("Wrong user/group")
        DIRAC.exit(2)

  if userGroup:
    from DIRAC.Core.Security.ProxyInfo import getProxyInfo
    res = getProxyInfo()
    if not res['OK']:
      gLogger.fatal("Can't get proxy info", res['Message'])
      DIRAC.exit(1)
    properties = res['Value'].get('groupProperties', [])
    if 'FileCatalogManagement' not in properties:
      gLogger.error("You need to use a proxy from a group with FileCatalogManagement")
      DIRAC.exit(5)

  from LHCbDIRAC.DataManagementSystem.Utilities.FCUtilities import chown
  from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import addFilesToTransformation
  from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  trClient = TransformationClient()
  transList = getTransformations(Script.getPositionalArgs())
  if not transList:
    DIRAC.exit(1)

  requestedLFNs = dmScript.getOption('LFNs')
  if not requestedLFNs:
    gLogger.always('No files to add')
    DIRAC.exit(0)

  if userGroup:
    directories = set([os.path.dirname(lfn) for lfn in requestedLFNs])
    res = chown(directories, user=userGroup[0], group=userGroup[1])
    if not res['OK']:
      gLogger.fatal("Error changing ownership", res['Message'])
      DIRAC.exit(3)
    gLogger.notice("Successfully changed owner/group for %d directories" % res['Value'])

  rc = 0
  for transID in transList:
    res = addFilesToTransformation(transID, requestedLFNs, addRunInfo=runInfo)
    if res['OK']:
      gLogger.always('Successfully added %d files to transformation %d' % (len(res['Value']), transID))
    else:
      gLogger.always('Failed to add %d files to transformation %d' % (len(requestedLFNs), transID), res['Message'])
      rc = 2
  DIRAC.exit(rc)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_RSS_Policy_Configurations."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

import LHCbDIRAC.ResourceStatusSystem.Policy.Configurations as moduleTested

__RCSID__ = "$Id$"

################################################################################


class Configurations_TestCase(unittest.TestCase):

  def setUp(self):
    """Setup."""

    self.moduleTested = moduleTested

  def tearDown(self):
    """Tear down."""

    del self.moduleTested

################################################################################


class Configurations_Success(Configurations_TestCase):

  def test_PoliciesMetaLHCb(self):

    policies = self.moduleTested.POLICIESMETA_LHCB

    policyKeys = set(['description', 'module', 'command', 'args'])

    for policyName, policy in policies.items():
      print(policyName)
      self.assertEqual(policyKeys, set(policy))
      self.assertTrue(isinstance(policy['description'], str))
      self.assertTrue(isinstance(policy['module'], str))
      self.assertTrue(isinstance(policy['command'], (tuple, None)))
      self.assertTrue(isinstance(policy['args'], (dict, None)))


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(Configurations_TestCase)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(Configurations_Success))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This is the Data Integrity Client which allows the simple reporting of
problematic file and replicas to the IntegrityDB and their status correctly
updated in the FileCatalog."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import six

from DIRAC import S_OK, gLogger
from DIRAC.Core.Utilities.Adler import compareAdler
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient as DIRACDataIntegrityClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks

__RCSID__ = "$Id$"


class DataIntegrityClient(DIRACDataIntegrityClient):

  def __init__(self):
    """Extending DIRAC's DIRACDataIntegrityClient init."""
    super(DataIntegrityClient, self).__init__()

    self.cc = ConsistencyChecks()

  ##########################################################################
  #
  # This section contains the specific methods for BK->FC checks
  #

  def productionToCatalog(self, productionID):
    """This obtains the file information from the BK and checks these files are
    present in the FC."""
    gLogger.info("-" * 40)
    gLogger.info("Performing the BK->FC check")
    gLogger.info("-" * 40)
    res = self.__getProductionFiles(productionID)
    if not res['OK']:
      return res
    noReplicaFiles = res['Value']['GotReplicaNo']
    yesReplicaFiles = res['Value']['GotReplicaYes']
    # For the files marked as existing we perfom catalog check
    res = self.cc._getCatalogMetadata(yesReplicaFiles)
    if not res['OK']:
      return res
    catalogMetadata, missingCatalogFiles, zeroSizeFiles = res['Value']
    if missingCatalogFiles:
      self._reportProblematicFiles(missingCatalogFiles, 'LFNCatalogMissing')
    if zeroSizeFiles:
      self._reportProblematicFiles(zeroSizeFiles, 'LFNZeroSize')

    # Try and get the metadata for files that shouldn't exist in the catalog
    if noReplicaFiles:
      res = self.__checkCatalogForBKNoReplicas(noReplicaFiles)
      if not res['OK']:
        return res
      catalogMetadata.update(res['Value'])
    # Get the replicas for the files found to exist in the catalog
    res = self.cc._getCatalogReplicas(list(catalogMetadata))
    if not res['OK']:
      return res
    replicas, zeroReplicaFiles = res['Value']
    if zeroReplicaFiles:
      self._reportProblematicFiles(zeroReplicaFiles, 'LFNZeroReplicas')
    resDict = {'CatalogMetadata': catalogMetadata, 'CatalogReplicas': replicas}
    return S_OK(resDict)

  def __checkCatalogForBKNoReplicas(self, lfns):
    """Checks the catalog existence for given files."""
    gLogger.info('Checking the catalog existence of %s files' % len(lfns))

    res = self.fc.getFileMetadata(lfns)
    if not res['OK']:
      gLogger.error('Failed to get catalog metadata', res['Message'])
      return res
    allMetadata = res['Value']['Successful']
    existingCatalogFiles = list(allMetadata)
    if existingCatalogFiles:
      self._reportProblematicFiles(existingCatalogFiles, 'BKReplicaNo')
    gLogger.info('Checking the catalog existence of files complete')
    return S_OK(allMetadata)

  def __getProductionFiles(self, productionID):
    """This method queries the bookkeeping and obtains the file metadata for
    the given production."""
    gLogger.info("Attempting to get files for production %s" % productionID)
    res = BookkeepingClient().getProductionFiles(productionID, 'ALL')
    if not res['OK']:
      return res
    yesReplicaFiles = []
    noReplicaFiles = []
    badReplicaFiles = []
    badBKFileSize = []
    badBKGUID = []
    allMetadata = res['Value']
    gLogger.info("Obtained at total of %s files" % len(allMetadata))
    totalSize = 0
    for lfn, bkMetadata in allMetadata.items():
      if bkMetadata['FileType'] != 'LOG':
        if bkMetadata['GotReplica'] == 'Yes':
          yesReplicaFiles.append(lfn)
          if bkMetadata['FileSize']:
            totalSize += int(bkMetadata['FileSize'])
        elif bkMetadata['GotReplica'] == 'No':
          noReplicaFiles.append(lfn)
        else:
          badReplicaFiles.append(lfn)
        if not bkMetadata['FileSize']:
          badBKFileSize.append(lfn)
        if not bkMetadata['GUID']:
          badBKGUID.append(lfn)
    if badReplicaFiles:
      self._reportProblematicFiles(badReplicaFiles, 'BKReplicaBad')
    if badBKFileSize:
      self._reportProblematicFiles(badBKFileSize, 'BKSizeBad')
    if badBKGUID:
      self._reportProblematicFiles(badBKGUID, 'BKGUIDBad')
    gLogger.info("%s files marked with replicas with total size %s bytes" % (len(yesReplicaFiles), totalSize))
    gLogger.info("%s files marked without replicas" % len(noReplicaFiles))
    resDict = {'BKMetadata': allMetadata, 'GotReplicaYes': yesReplicaFiles, 'GotReplicaNo': noReplicaFiles}
    return S_OK(resDict)

  ##########################################################################
  #
  # This section contains the specific methods for FC->BK checks
  #

  def catalogDirectoryToBK(self, lfnDir):
    """This obtains the replica and metadata information from the catalog for
    the supplied directory and checks against the BK."""
    gLogger.info("-" * 40)
    gLogger.info("Performing the FC->BK check")
    gLogger.info("-" * 40)
    if isinstance(lfnDir, six.string_types):
      lfnDir = [lfnDir]
    res = self.__getCatalogDirectoryContents(lfnDir)
    if not res['OK']:
      return res
    replicas = res['Value']['Replicas']
    catalogMetadata = res['Value']['Metadata']
    resDict = {'CatalogMetadata': catalogMetadata, 'CatalogReplicas': replicas}
    if not catalogMetadata:
      gLogger.warn('No files found in directory %s' % lfnDir)
      return S_OK(resDict)
    lfns = []
    for repDict in replicas:
      lfns.append(repDict)
    missingLFNs, noFlagLFNs, _okLFNs = self.cc._getBKMetadata(lfns)
    if missingLFNs:
      self._reportProblematicFiles(missingLFNs, 'LFNBKMissing')
    if noFlagLFNs:
      self._reportProblematicFiles(noFlagLFNs, 'BKReplicaNo')
    return S_OK(resDict)

  def catalogFileToBK(self, lfns):
    """This obtains the replica and metadata information from the catalog and
    checks against the storage elements."""
    gLogger.info("-" * 40)
    gLogger.info("Performing the FC->BK check")
    gLogger.info("-" * 40)
    if isinstance(lfns, six.string_types):
      lfns = [lfns]

    res = self.cc._getCatalogMetadata(lfns)
    if not res['OK']:
      return res
    catalogMetadata, missingCatalogFiles, zeroSizeFiles = res['Value']
    if missingCatalogFiles:
      self._reportProblematicFiles(missingCatalogFiles, 'LFNCatalogMissing')
    if zeroSizeFiles:
      self._reportProblematicFiles(zeroSizeFiles, 'LFNZeroSize')

    res = self.cc._getCatalogReplicas(list(catalogMetadata))
    if not res['OK']:
      return res
    replicas, _zeroReplicaFiles = res['Value']

    lfns = []
    for repDict in replicas:
      lfns.append(repDict)
    missingLFNs, noFlagLFNs, _okLFNs = self.cc._getBKMetadata(lfns)
    if missingLFNs:
      self._reportProblematicFiles(missingLFNs, 'LFNBKMissing')
    if noFlagLFNs:
      self._reportProblematicFiles(noFlagLFNs, 'BKReplicaNo')

    resDict = {'CatalogMetadata': catalogMetadata, 'CatalogReplicas': replicas}
    return S_OK(resDict)

  ##########################################################################
  #
  # This section contains the resolution methods for various prognoses
  #

  def resolveBKReplicaYes(self, problematicDict):
    """This takes the problematic dictionary returned by the integrity DB and
    resolved the BKReplicaYes prognosis."""
    lfn = problematicDict['LFN']
    fileID = problematicDict['FileID']

    res = returnSingleResult(self.fc.exists(lfn))
    if not res['OK']:
      return self.__returnProblematicError(fileID, res)
    removeBKFile = False
    # If the file does not exist in the catalog
    if not res['Value']:
      gLogger.info("BKReplicaYes file (%d) does not exist in the catalog. Removing..." % fileID)
      removeBKFile = True
    else:
      gLogger.info("BKReplicaYes file (%d) found to exist in the catalog" % fileID)
      # If the file has no replicas in the catalog
      res = returnSingleResult(self.fc.getReplicas(lfn))
      if (not res['OK']) and (res['Message'] == 'File has zero replicas'):
        gLogger.info("BKReplicaYes file (%d) found to exist without replicas. Removing..." % fileID)
        removeBKFile = True
    if removeBKFile:
      # Remove the file from the BK because it does not exist
      res = returnSingleResult(FileCatalog(catalogs=['BookkeepingDB']).removeFile(lfn))
      if not res['OK']:
        return self.__returnProblematicError(fileID, res)
      gLogger.info("BKReplicaYes file (%d) removed from bookkeeping" % fileID)
    return self.__updateCompletedFiles('BKReplicaYes', fileID)

  def resolveBKReplicaNo(self, problematicDict):
    """This takes the problematic dictionary returned by the integrity DB and
    resolved the BKReplicaNo prognosis."""
    lfn = problematicDict['LFN']
    fileID = problematicDict['FileID']

    res = returnSingleResult(self.fc.exists(lfn))
    if not res['OK']:
      return self.__returnProblematicError(fileID, res)
    # If the file exists in the catalog
    if not res['Value']:
      return self.__updateCompletedFiles('BKReplicaNo', fileID)
    gLogger.info("BKReplicaNo file (%d) found to exist in the catalog" % fileID)
    # and has available replicas
    res = returnSingleResult(self.fc.getCatalogReplicas(lfn))
    if not res['OK']:
      return self.__returnProblematicError(fileID, res)
    if not res['Value']:
      gLogger.info("BKReplicaNo file (%d) found to have no replicas" % fileID)
      return self.changeProblematicPrognosis(fileID, 'LFNZeroReplicas')
    gLogger.info("BKReplicaNo file (%d) found to have replicas" % fileID)
    res = returnSingleResult(FileCatalog(catalogs=['BookkeepingDB']).addFile(lfn))
    if not res['OK']:
      return self.__returnProblematicError(fileID, res)
    return self.__updateCompletedFiles('BKReplicaNo', fileID)

  def checkPhysicalFiles(self, replicas, catalogMetadata, ses=[], fixIt=False):
    """This obtains takes the supplied replica and metadata information
    obtained from the catalog and checks against the storage elements."""
    gLogger.info("-" * 40)
    gLogger.info("Performing the FC->SE check")
    gLogger.info("-" * 40)

    seLfns = {}
    for lfn, replicaDict in replicas.items():
      for se in replicaDict:
        if (ses) and (se not in ses):
          continue
        seLfns.setdefault(se, []).append(lfn)
    gLogger.info('%s %s' % ('Storage Element'.ljust(20), 'Replicas'.rjust(20)))

    for se in sorted(seLfns):
      lfns = seLfns[se]

      sizeMismatch = []
      checksumMismatch = []
      checksumBadInFC = []
      res = self.__checkPhysicalFileMetadata(lfns, se)
      if not res['OK']:
        gLogger.error('Failed to get physical file metadata.', res['Message'])
        return res
      for lfn, metadata in res['Value'].items():
        if lfn in catalogMetadata:
          if (metadata['Size'] != catalogMetadata[lfn]['Size']) and (metadata['Size'] != 0):
            sizeMismatch.append((lfn, 'deprecatedUrl', se, 'CatalogPFNSizeMismatch'))
          if not compareAdler(metadata['Checksum'], catalogMetadata[lfn]['Checksum']):
            if compareAdler(metadata['Checksum'].replace('x', '0'), catalogMetadata[lfn]['Checksum'].replace('x', '0')):
              checksumBadInFC.append(
                  (lfn, 'deprecatedUrl', se, "%s %s" %
                   (metadata['Checksum'], catalogMetadata[lfn]['Checksum'])))
            else:
              checksumMismatch.append(
                  (lfn, 'deprecatedUrl', se, "%s %s" %
                   (metadata['Checksum'], catalogMetadata[lfn]['Checksum'])))
      if sizeMismatch:
        self.reportProblematicReplicas(sizeMismatch, se, 'CatalogPFNSizeMismatch', fixIt=fixIt)
      if checksumMismatch:
        self.reportProblematicReplicas(checksumMismatch, se, 'CatalogChecksumMismatch', fixIt=fixIt)
      if checksumBadInFC:
        self.reportProblematicReplicas(checksumBadInFC, se, 'CatalogChecksumToBeFixed', fixIt=fixIt)
    return S_OK()

  def __checkPhysicalFileMetadata(self, lfns, se):
    """Check obtain the physical file metadata and check the files are
    available."""
    gLogger.info('Checking the integrity of %s physical files at %s' % (len(lfns), se))

    res = StorageElement(se).getFileMetadata(lfns)
    if not res['OK']:
      gLogger.error('Failed to get metadata for lfns.', res['Message'])
      return res
    pfnMetadataDict = res['Value']['Successful']
    # If the replicas are completely missing
    missingReplicas = []
    for lfn, reason in res['Value']['Failed'].items():
      if re.search('File does not exist', reason):
        missingReplicas.append((lfn, 'deprecatedUrl', se, 'PFNMissing'))
    if missingReplicas:
      self.reportProblematicReplicas(missingReplicas, se, 'PFNMissing')
    lostReplicas = []
    unavailableReplicas = []
    zeroSizeReplicas = []
    # If the files are not accessible
    for lfn, metadata in pfnMetadataDict.items():
      if metadata.get('Lost', False):
        lostReplicas.append((lfn, se, 'PFNLost'))
      if metadata.get('Unavailable', not metadata['Accessible']):
        unavailableReplicas.append((lfn, 'deprecatedUrl', se, 'PFNUnavailable'))
      if metadata['Size'] == 0:
        zeroSizeReplicas.append((lfn, 'deprecatedUrl', se, 'PFNZeroSize'))
    if lostReplicas:
      self.reportProblematicReplicas(lostReplicas, se, 'PFNLost')
    if unavailableReplicas:
      self.reportProblematicReplicas(unavailableReplicas, se, 'PFNUnavailable')
    if zeroSizeReplicas:
      self.reportProblematicReplicas(zeroSizeReplicas, se, 'PFNZeroSize')
    gLogger.info('Checking the integrity of physical files at %s complete' % se)
    return S_OK(pfnMetadataDict)

  def reportProblematicReplicas(self, replicaTuple, se, reason, fixIt=False):
    """Simple wrapper function around setReplicaProblematic."""
    gLogger.info('The following %s files had %s at %s' % (len(replicaTuple), reason, se))
    for lfn, pfn, se, reason1 in sorted(replicaTuple):
      if reason1 == reason:
        reason1 = ''
      if lfn:
        gLogger.info(lfn, reason1)
      else:
        gLogger.info(pfn, reason1)
    if fixIt:
      res = self.setReplicaProblematic(replicaTuple, sourceComponent='DataIntegrityClient')
      if not res['OK']:
        gLogger.info('Failed to update integrity DB with replicas', res['Message'])
      else:
        gLogger.info('Successfully updated integrity DB with replicas')

  ##########################################################################
  #
  # This section contains the specific methods for obtaining replica and metadata information from the catalog
  #

  def __getCatalogDirectoryContents(self, lfnDir):
    """Obtain the contents of the supplied directory."""
    gLogger.info('Obtaining the catalog contents for %s directories' % len(lfnDir))

    activeDirs = list(lfnDir)
    allFiles = {}
    while len(activeDirs) > 0:
      currentDir = activeDirs[0]
      res = self.fc.listDirectory(currentDir, verbose=True)
      activeDirs.remove(currentDir)
      if not res['OK']:
        gLogger.error('Failed to get directory contents', res['Message'])
        return res
      elif currentDir in res['Value']['Failed']:
        gLogger.error('Failed to get directory contents', '%s %s' % (currentDir, res['Value']['Failed'][currentDir]))
      else:
        dirContents = res['Value']['Successful'][currentDir]
        activeDirs.extend(dirContents['SubDirs'])
        allFiles.update(dirContents['Files'])

    zeroReplicaFiles = []
    zeroSizeFiles = []
    allReplicaDict = {}
    allMetadataDict = {}
    for lfn, lfnDict in allFiles.items():
      lfnReplicas = {}
      for se, replicaDict in lfnDict['Replicas'].items():
        lfnReplicas[se] = replicaDict['PFN']
      if not lfnReplicas:
        zeroReplicaFiles.append(lfn)
      allReplicaDict[lfn] = lfnReplicas
      allMetadataDict[lfn] = lfnDict['MetaData']
      if lfnDict['MetaData']['Size'] == 0:
        zeroSizeFiles.append(lfn)
    if zeroReplicaFiles:
      self._reportProblematicFiles(zeroReplicaFiles, 'LFNZeroReplicas')
    if zeroSizeFiles:
      self._reportProblematicFiles(zeroSizeFiles, 'LFNZeroSize')
    gLogger.info('Obtained at total of %s files for the supplied directories' % len(allMetadataDict))
    resDict = {'Metadata': allMetadataDict, 'Replicas': allReplicaDict}
    return S_OK(resDict)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod:  WorkflowTaskAgent

.. module:  WorkflowTaskAgent

:synopsis:  Extension of the DIRAC WorkflowTaskAgent, to use LHCb clients.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.TransformationSystem.Agent.WorkflowTaskAgent import WorkflowTaskAgent as DIRACWorkflowTaskAgent

from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob
from LHCbDIRAC.TransformationSystem.Client.TaskManager import LHCbWorkflowTasks
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

AGENT_NAME = 'Transformation/WorkflowTaskAgent'


class WorkflowTaskAgent(DIRACWorkflowTaskAgent):
  """An AgentModule class to submit workflow tasks."""

  def __init__(self, *args, **kwargs):
    """c'tor."""
    DIRACWorkflowTaskAgent.__init__(self, *args, **kwargs)

  def _getClients(self):
    """LHCb clients."""
    res = DIRACWorkflowTaskAgent._getClients(self)

    outputDataModule = Operations().getValue("Transformations/OutputDataModule",
                                             "LHCbDIRAC.Core.Utilities.OutputDataPolicy")

    threadTransformationClient = TransformationClient()
    threadTaskManager = LHCbWorkflowTasks(outputDataModule=outputDataModule,
                                          jobClass=LHCbJob)
    res.update({'TransformationClient': threadTransformationClient,
                'TaskManager': threadTaskManager})
    return res



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Root Application Class."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities import DErrno

from LHCbDIRAC.Core.Utilities.RunApplication import RunApplication, LbRunError, LHCbApplicationError, LHCbDIRACError
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class RootApplication(ModuleBase):

  #############################################################################
  def __init__(self, bkClient=None, dm=None):
    """Module initialization."""

    self.log = gLogger.getSubLogger("RootApplication")
    super(RootApplication, self).__init__(self.log, bkClientIn=bkClient, dm=dm)

    self.version = __RCSID__

    self.applicationLog = ''
    self.applicationVersion = ''
    self.applicationName = ''
    self.rootType = ''
    self.arguments = ''
    self.systemConfig = ''

  #############################################################################
  def _resolveInputVariables(self):
    """By convention the workflow parameters are resolved here."""

    super(RootApplication, self)._resolveInputVariables()
    super(RootApplication, self)._resolveInputStep()

    if 'rootScript' in self.step_commons:
      self.applicationName = self.step_commons['rootScript']
    else:
      self.log.warn('No rootScript defined')

    if 'rootVersion' in self.step_commons:
      self.applicationVersion = self.step_commons['rootVersion']
    else:
      self.log.warn('No rootVersion defined')

    if 'rootType' in self.step_commons:
      self.rootType = self.step_commons['rootType']
    else:
      self.log.warn('No rootType defined')

    if 'arguments' in self.step_commons:
      self.arguments = self.step_commons['arguments']
      tmp = []
      for argument in self.arguments:
        if argument or not isinstance(argument, list):
          tmp.append(argument)
      self.arguments = tmp
    else:
      self.log.warn('No arguments specified')

    print(self.arguments)

  #############################################################################
  def execute(self, production_id=None, prod_job_id=None, wms_job_id=None,
              workflowStatus=None, stepStatus=None,
              wf_commons=None, step_commons=None,
              step_id=None, step_number=None):
    """The main execution method of the RootApplication module: runs a ROOT app
    using RunApplication module."""

    try:

      super(RootApplication, self).execute(self.version, production_id, prod_job_id, wms_job_id, workflowStatus,
                                           stepStatus, wf_commons, step_commons, step_number, step_id)

      self._resolveInputVariables()

      if not self.applicationVersion:
        raise RuntimeError('No Root Version defined')
      if not self.systemConfig:
        raise RuntimeError('No system configuration selected')
      if not self.applicationName:
        raise RuntimeError('No script defined')
      if not self.applicationLog:
        self.applicationLog = '%s.log' % self.applicationName
      if self.rootType.lower() not in ('c', 'py', 'bin', 'exe'):
        raise RuntimeError('Wrong root type defined')

      self.setApplicationStatus('Initializing RootApplication module')

      self.log.info("Executing application Root %s with CMT config %s " % (self.applicationVersion, self.systemConfig))

      if not os.path.exists(self.applicationName):
        self.log.error('rootScript not Found at %s' % os.path.abspath('.'))
        return S_ERROR('rootScript not Found')

      if self.rootType.lower() == 'c':
        rootCmd = ['root']
        rootCmd.append('-b')
        rootCmd.append('-f')
        if self.arguments:
          escapedArgs = []
          for arg in self.arguments:
            if isinstance(arg, str):
              escapedArgs.append('\'"%s"\'' % (arg))
            else:
              escapedArgs.append('%s' % (arg))

          macroArgs = r'%s\(%s\)' % (self.applicationName, ','.join(escapedArgs))
          rootCmd.append(macroArgs)
        else:
          rootCmd.append(self.applicationName)

      elif self.rootType.lower() == 'py':

        rootCmd = ['python']
        rootCmd = [os.path.abspath(self.applicationName)]
        if self.arguments:
          rootCmd += self.arguments

      if os.path.exists(self.applicationLog):
        os.remove(self.applicationLog)

      # How to run the application
      ra = RunApplication()
      # lb-run stuff
      ra.applicationName = self.applicationName
      ra.applicationVersion = self.applicationVersion
      ra.systemConfig = self.systemConfig
      # actual stuff to run
      ra.command = rootCmd
      ra.applicationLog = self.applicationLog
      ra.numberOfProcessors = self.numberOfProcessors

      # Now really running
      ra.run()  # This would trigger an exception in case of failure, or application status != 0

      # Return OK assuming that subsequent module will spot problems
      self.setApplicationStatus('%s (Root) Successful' % self.applicationName)
      return S_OK()

    except LbRunError as lbre:  # This is the case for lb-run/environment errors
      self.setApplicationStatus(repr(lbre))
      return S_ERROR(DErrno.EWMSRESC, str(lbre))
    except LHCbApplicationError as lbae:  # This is the case for real application errors
      self.setApplicationStatus(repr(lbae))
      return S_ERROR(str(lbae))
    except LHCbDIRACError as lbde:  # This is the case for LHCbDIRAC errors (e.g. subProcess call failed)
      self.setApplicationStatus(repr(lbde))
      return S_ERROR(str(lbde))
    except Exception as e:  # pylint:disable=broad-except
      self.log.exception("Failure in RootApplication execute module", lException=e)
      return S_ERROR("Error in RootApplication module")

    finally:
      super(RootApplication, self).finalize(self.version)
